import pytest

from grunnur import API
from grunnur.api import all_api_ids


def test_all(mock_backend_factory):
    api_id = all_api_ids()[0]
    mock_backend_factory.mock(api_id)

    apis = API.all_available()
    assert len(apis) == 1
    assert apis[0].id == api_id


def test_all_by_shortcut(mock_backend):
    api_id = mock_backend.api_id
    mock_backend.add_devices(["Device1", "Device2"])

    apis = API.all_by_shortcut(api_id.shortcut)
    assert len(apis) == 1
    assert apis[0].id == api_id


def test_all_by_shortcut_none(mock_backend_factory):
    api_ids = all_api_ids()
    for api_id in api_ids:
        backend = mock_backend_factory.mock(api_id)
        backend.add_devices(["Device1", "Device2"])

    apis = API.all_by_shortcut()
    assert len(apis) == len(api_ids)
    assert {api.id for api in apis} == set(api_ids)


@pytest.mark.usefixtures("mock_backend_factory")
def test_all_by_shortcut_not_available():
    api_id = all_api_ids()[0]
    with pytest.raises(ValueError, match="cuda API is not available"):
        API.all_by_shortcut(api_id.shortcut)


def test_all_by_shortcut_not_found():
    with pytest.raises(ValueError, match="Invalid API shortcut: something non-existent"):
        API.all_by_shortcut("something non-existent")


def test_from_api_id(mock_backend_factory):
    for api_id in all_api_ids():
        with pytest.raises(ImportError):
            API.from_api_id(api_id)

        mock_backend_factory.mock(api_id)
        api = API.from_api_id(api_id)
        assert api.id == api_id


def test_eq(mock_backend_factory):
    api_id0 = all_api_ids()[0]
    api_id1 = all_api_ids()[1]

    mock_backend_factory.mock(api_id0)
    mock_backend_factory.mock(api_id1)

    api0_v1 = API.from_api_id(api_id0)
    api0_v2 = API.from_api_id(api_id0)
    api1 = API.from_api_id(api_id1)

    assert api0_v1 is not api0_v2
    assert api0_v1 == api0_v2
    assert api0_v1 != api1


def test_hash(mock_backend_factory):
    api_id0 = all_api_ids()[0]
    api_id1 = all_api_ids()[1]

    mock_backend_factory.mock(api_id0)
    mock_backend_factory.mock(api_id1)

    api0 = API.from_api_id(api_id0)
    api1 = API.from_api_id(api_id1)

    d = {api0: 0, api1: 1}
    assert d[api0] == 0
    assert d[api1] == 1


def test_getitem(mock_backend_pyopencl):
    api_id = mock_backend_pyopencl.api_id

    mock_backend_pyopencl.add_platform_with_devices("Platform0", ["Device0"])
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1"])

    api = API.from_api_id(api_id)
    assert api.platforms[0].name == "Platform0"
    assert api.platforms[1].name == "Platform1"


def test_attributes(mock_backend):
    api = API.from_api_id(mock_backend.api_id)
    assert str(mock_backend.api_id) == "id(" + api.shortcut + ")"
    assert api.id == mock_backend.api_id
    assert api.shortcut == mock_backend.api_id.shortcut
    assert str(api) == "api(" + api.shortcut + ")"



import os.path
import re
from pathlib import Path
from typing import NamedTuple

import numpy
import pytest

from grunnur import (
    API,
    Array,
    Buffer,
    CompilationError,
    Context,
    DeviceFilter,
    MultiArray,
    MultiQueue,
    Program,
    Queue,
    StaticKernel,
    cuda_api_id,
    opencl_api_id,
)
from grunnur.template import DefTemplate, Template
from grunnur.testing import MockDefTemplate, MockKernel, PyCUDADeviceInfo

SRC_OPENCL = """
__kernel void multiply(__global int *dest, __global int *a, __global int *b, int c)
{
    const int i = get_global_id(0);
    dest[i] = a[i] * b[i] + c;
}
"""

SRC_CUDA = """
extern "C" __global__ void multiply(int *dest, int *a, int *b, int c)
{
    const int i = threadIdx.x + blockIdx.x * blockDim.x;
    dest[i] = a[i] * b[i] + c;
}
"""

SRC_GENERIC = """
KERNEL void multiply(GLOBAL_MEM int *dest, GLOBAL_MEM int *a, GLOBAL_MEM int *b, int c)
{
    const int i = get_global_id(0);
    dest[i] = a[i] * b[i] + c;
}
"""


@pytest.mark.parametrize("no_prelude", [False, True], ids=["with_prelude", "no_prelude"])
def test_compile(mock_or_real_context, no_prelude):
    context, mocked = mock_or_real_context

    if mocked:
        src = MockDefTemplate(kernels=[MockKernel("multiply", [None, None, None, numpy.int32])])
    elif no_prelude:
        src = SRC_CUDA if context.api.id == cuda_api_id() else SRC_OPENCL
    else:
        src = SRC_GENERIC

    program = Program([context.device], src, no_prelude=no_prelude)

    if mocked and no_prelude:
        assert program.sources[context.device].prelude.strip() == ""

    length = 64

    a = numpy.arange(length).astype(numpy.int32)
    b = numpy.arange(length).astype(numpy.int32) + 1
    c = numpy.int32(3)
    ref = a * b + c

    queue = Queue(context.device)

    a_dev = Array.from_host(queue, a)
    b_dev = Array.from_host(queue, b)

    res_dev = Array.empty(context.device, [length], numpy.int32)
    # Check that passing both Arrays and Buffers is supported
    # Pass one of the buffers as a subregion, too.
    a_dev_view = a_dev.data.get_sub_region(0, a_dev.data.size)
    program.kernel.multiply(queue, [length], None, res_dev, a_dev_view, b_dev.data, c)
    res = res_dev.get(queue)
    if not mocked:
        assert (res == ref).all()

    # Explicit local_size
    res2_dev = Array.from_host(queue, a)  # Array.empty(queue, length, numpy.int32)
    program.kernel.multiply(queue, [length], [length // 2], res2_dev, a_dev, b_dev, c)
    res2 = res2_dev.get(queue)
    if not mocked:
        assert (res2 == ref).all()


def test_compile_multi_device(mock_or_real_multi_device_context):
    context, mocked = mock_or_real_multi_device_context
    devices = context.devices[[1, 0]]

    if mocked:
        src = MockDefTemplate(kernels=[MockKernel("multiply", [None, None, None, numpy.int32])])
    else:
        src = SRC_GENERIC

    length = 64

    program = Program(devices, src)
    a = numpy.arange(length).astype(numpy.int32)
    b = numpy.arange(length).astype(numpy.int32) + 1
    c = numpy.int32(3)
    ref = a * b + c

    mqueue = MultiQueue.on_devices(devices)

    a_dev = MultiArray.from_host(mqueue, a)
    b_dev = MultiArray.from_host(mqueue, b)

    res_dev = MultiArray.empty(devices, [length], numpy.int32)
    program.kernel.multiply(mqueue, a_dev.shapes, None, res_dev, a_dev, b_dev, c)
    res = res_dev.get(mqueue)
    if not mocked:
        assert (res == ref).all()

    # Test argument unpacking from dictionaries
    res_dev = MultiArray.empty(devices, [length], numpy.int32)
    program.kernel.multiply(
        mqueue,
        a_dev.shapes,
        {device: None for device in devices},
        res_dev,
        a_dev.subarrays,
        b_dev,
        c,
    )
    res = res_dev.get(mqueue)
    if not mocked:
        assert (res == ref).all()


def test_mismatched_devices(mock_4_device_context):
    context = mock_4_device_context
    src = MockDefTemplate(kernels=[MockKernel("multiply", [None, None, None, numpy.int32])])
    program = Program(context.devices, src)
    with pytest.raises(ValueError, match="Mismatched device sets for global and local sizes"):
        program.kernel.multiply.prepare(
            {context.devices[0]: [10], context.devices[2]: [20]},
            {context.devices[0]: None, context.devices[1]: None},
        )


SRC_CONSTANT_MEM = """
KERNEL void copy_from_cm(
    GLOBAL_MEM int *dest
#ifdef GRUNNUR_OPENCL_API
    , CONSTANT_MEM int *cm1
    , CONSTANT_MEM int *cm2
    , CONSTANT_MEM int *cm3
#endif
    )
{
    const int i = get_global_id(0);
    dest[i] = cm1[i] + cm2[i] + cm3[i];
}
"""


SRC_CONSTANT_MEM_STATIC = """
KERNEL void copy_from_cm(
    GLOBAL_MEM int *dest
#ifdef GRUNNUR_OPENCL_API
    , CONSTANT_MEM int *cm1
    , CONSTANT_MEM int *cm2
    , CONSTANT_MEM int *cm3
#endif
    )
{
    if (${static.skip}()) return;
    const int i = ${static.global_id}(0);
    dest[i] = cm1[i] + cm2[i] + cm3[i];
}
"""


def _test_constant_memory(context, mocked, is_static):
    cm1 = numpy.arange(16).astype(numpy.int32)
    cm2 = numpy.arange(16).astype(numpy.int32) * 2 + 1
    cm3 = numpy.arange(16).astype(numpy.int32) * 3 + 2

    if mocked:
        kernel = MockKernel(
            "copy_from_cm",
            [None] if context.api.id == cuda_api_id() else [None, None, None, None],
            max_total_local_sizes={0: 1024},
        )
        src = MockDefTemplate(
            constant_mem={
                "cm1": cm1.size * cm1.dtype.itemsize,
                "cm2": cm2.size * cm2.dtype.itemsize,
                "cm3": cm3.size * cm3.dtype.itemsize,
            },
            kernels=[kernel],
        )
    else:
        src = SRC_CONSTANT_MEM_STATIC if is_static else SRC_CONSTANT_MEM

    queue = Queue(context.device)

    cm1_dev = Array.from_host(queue, cm1)
    cm2_dev = Array.from_host(queue, cm2)
    cm3_dev = Array.from_host(queue, cm3)
    res_dev = Array.empty(context.device, [16], numpy.int32)

    class Metadata(NamedTuple):
        shape: tuple[int, ...]
        dtype: numpy.dtype

    if context.api.id == cuda_api_id():
        # Use different forms of constant array representation
        constant_arrays = dict(
            cm1=cm1,  # as an array(-like) object
            cm2=Metadata(cm2.shape, cm2.dtype),  # as a metadata-like
            cm3=cm3_dev,
        )  # as a device array

        if is_static:
            copy_from_cm = StaticKernel(
                [context.device],
                src,
                "copy_from_cm",
                global_size=[16],
                constant_arrays=constant_arrays,
            )
            copy_from_cm.set_constant_array(queue, "cm1", cm1_dev)  # setting from a device array
            copy_from_cm.set_constant_array(queue, "cm2", cm2)  # setting from a host array
            copy_from_cm.set_constant_array(
                queue, "cm3", cm3_dev.data
            )  # setting from a host buffer
        else:
            program = Program([context.device], src, constant_arrays=constant_arrays)
            program.set_constant_array(queue, "cm1", cm1_dev)  # setting from a device array
            program.set_constant_array(queue, "cm2", cm2)  # setting from a host array
            program.set_constant_array(queue, "cm3", cm3_dev.data)  # setting from a host buffer
            copy_from_cm = lambda queue, *args: program.kernel.copy_from_cm(
                queue, [16], None, *args
            )

        copy_from_cm(queue, res_dev)
    else:
        if is_static:
            copy_from_cm = StaticKernel([context.device], src, "copy_from_cm", global_size=[16])
        else:
            program = Program([context.device], src)
            copy_from_cm = lambda queue, *args: program.kernel.copy_from_cm(
                queue, [16], None, *args
            )

        copy_from_cm(queue, res_dev, cm1_dev, cm2_dev, cm3_dev)

    res = res_dev.get(queue)

    if not mocked:
        assert (res == cm1 + cm2 + cm3).all()


def test_constant_memory(mock_or_real_context):
    context, mocked = mock_or_real_context
    _test_constant_memory(context=context, mocked=mocked, is_static=False)


SRC_COMPILE_ERROR = """
KERNEL void compile_error(GLOBAL_MEM int *dest)
{
    const int i = get_global_id(0);
    dest[i] = 1;
    zzz
}
"""


def test_compilation_error(mock_or_real_context, capsys):
    context, mocked = mock_or_real_context

    src = MockDefTemplate(should_fail=True) if mocked else SRC_COMPILE_ERROR

    with pytest.raises(CompilationError):
        Program([context.device], src)

    captured = capsys.readouterr()
    assert "Failed to compile on device" in captured.out

    # check that the full source is shown (including the prelude)
    assert "#define GRUNNUR_" in captured.out

    if mocked:
        assert "<<< mock source >>>" in captured.out
    else:
        assert "KERNEL void compile_error(GLOBAL_MEM int *dest)" in captured.out


def test_keep(mock_or_real_context, capsys):
    context, mocked = mock_or_real_context

    if mocked:
        src = MockDefTemplate(kernels=[MockKernel("multiply", [None, None, None, numpy.int32])])
    else:
        src = SRC_GENERIC

    Program([context.device], src, keep=True)
    captured = capsys.readouterr()
    path = Path(re.match(r"\*\*\* compiler output in (.*)", captured.out).group(1))
    assert path.is_dir()

    if context.api.id == opencl_api_id():
        srcfile = path / "kernel.cl"
    elif context.api.id == cuda_api_id():
        srcfile = path / "kernel.cu"

    with srcfile.open() as f:
        source = f.read()

    assert str(src) in source


def test_compiler_options(mock_context):
    src = MockDefTemplate(kernels=[MockKernel("multiply", [None, None, None, numpy.int32])])
    program = Program([mock_context.device], src, compiler_options=["--my_option"])

    adapter = program._sd_programs[mock_context.device]._sd_program_adapter

    if mock_context.api.id == opencl_api_id():
        assert "--my_option" in adapter._pyopencl_program.test_get_options()
    elif mock_context.api.id == cuda_api_id():
        assert "--my_option" in adapter._pycuda_program.test_get_options()


def test_wrong_kernel_argument(mock_context):
    src = MockDefTemplate(kernels=[MockKernel("multiply", [None, None, None, numpy.int32])])
    program = Program([mock_context.device], src)
    queue = Queue(mock_context.device)

    buf = Buffer.allocate(mock_context.device, 100)

    with pytest.raises(TypeError, match="Incorrect argument type: <class 'int'>"):
        program.kernel.multiply(queue, (100,), None, buf, buf, buf, 4)

    if mock_context.api.id == cuda_api_id():
        with pytest.raises(
            RuntimeError, match="A previously freed allocation or an incorrect address is used"
        ):
            program.kernel.multiply(queue, (100,), None, numpy.uintp(0), buf, buf, 4)


def test_wrong_device_idxs(mock_4_device_context):
    src = MockDefTemplate(kernels=[MockKernel("multiply", [None])])

    context = mock_4_device_context
    program = Program(context.devices[[0, 1]], src)
    mqueue = MultiQueue.on_devices(context.devices[[2, 1]])
    res_dev = MultiArray.empty(context.devices[[2, 1]], [16], numpy.int32)

    # Using all the queue's devices (1, 2)
    with pytest.raises(ValueError, match="Requested execution on devices"):
        program.kernel.multiply(mqueue, [8], None, res_dev)


def test_wrong_context(mock_backend):
    mock_backend.add_devices(["Device0"])

    src = MockDefTemplate(kernels=[MockKernel("multiply", [None])])

    api = API.from_api_id(mock_backend.api_id)
    context = Context.from_devices([api.platforms[0].devices[0]])
    context2 = Context.from_devices([api.platforms[0].devices[0]])

    res_dev = Array.empty(context.device, [16], numpy.int32)

    program = Program([context.device], src)
    queue = Queue(context2.device)

    with pytest.raises(
        ValueError, match="The provided queue must belong to the same context this program uses"
    ):
        program.kernel.multiply(queue, [8], None, res_dev)

    # If we don't do anything here, contexts may be deactivated in an incorrect order
    # (on drop of the corresponding `Context` objects).
    # Grunnur cannot resolve this, so we need to deactivate the currently active context
    # (the second one) manually, and the first one will be deactivated on drop.

    context2.deactivate()


def test_set_constant_array_errors(mock_4_device_context):
    context = mock_4_device_context

    api = API.from_api_id(mock_4_device_context.api.id)
    other_context = Context.from_criteria(api)
    other_queue = Queue(other_context.devices[0])
    # Contexts don't know about each other and can't interact with stack in a consistent manner.
    # So we deactivate the other context if we're on CUDA API.
    if api.id == cuda_api_id():
        other_context.deactivate()

    cm1 = numpy.arange(16).astype(numpy.int32)
    src = MockDefTemplate(
        kernels=[
            MockKernel("kernel", [], max_total_local_sizes={0: 1024, 1: 1024, 2: 1024, 3: 1024})
        ],
        constant_mem={"cm1": cm1.size * cm1.dtype.itemsize},
    )
    queue = Queue(context.devices[0])

    if context.api.id == cuda_api_id():
        program = Program(context.devices, src, constant_arrays=dict(cm1=cm1))

        with pytest.raises(
            ValueError,
            match="The provided queue must belong to the same context as this program uses",
        ):
            program.set_constant_array(other_queue, "cm1", cm1)

        with pytest.raises(TypeError, match="Unsupported array type"):
            program.set_constant_array(queue, "cm1", [1])

        with pytest.raises(ValueError, match="Incorrect size of the constant buffer;"):
            program.set_constant_array(queue, "cm1", cm1[:8])

        with pytest.raises(TypeError, match="Unknown constant array metadata type"):
            program = Program(context.devices[[0, 1, 2]], src, constant_arrays=dict(cm1=1))

        program = Program(context.devices[[0, 1, 2]], src, constant_arrays=dict(cm1=cm1))
        queue3 = Queue(context.devices[3])

        with pytest.raises(
            ValueError, match="The program was not compiled for the device this queue uses"
        ):
            program.set_constant_array(queue3, "cm1", cm1)

    else:
        with pytest.raises(
            ValueError, match="Compile-time constant arrays are only supported for CUDA API"
        ):
            program = Program(context.devices, src, constant_arrays=dict(cm1=cm1))

        program = Program(context.devices, src)
        with pytest.raises(
            RuntimeError, match="OpenCL does not allow setting constant arrays externally"
        ):
            program.set_constant_array(queue, "cm1", cm1)

        with pytest.raises(
            ValueError, match="Compile-time constant arrays are only supported for CUDA API"
        ):
            sk = StaticKernel(context.devices, src, "kernel", [1024], constant_arrays=dict(cm1=cm1))

        sk = StaticKernel(context.devices, src, "kernel", [1024])
        with pytest.raises(
            RuntimeError, match="OpenCL does not allow setting constant arrays externally"
        ):
            sk.set_constant_array(queue, "cm1", cm1)


def test_max_total_local_sizes(mock_backend):
    mock_backend.add_devices(["Device1", "Device2 - tag", "Device3 - tag", "Device4"])
    api = API.from_api_id(mock_backend.api_id)
    context = Context.from_criteria(
        api, devices_num=2, device_filter=DeviceFilter(include_masks=["tag"])
    )

    # Providing max_total_local_sizes for all possible devices to make sure
    # only the ones corresponding to the context will get picked up
    kernel = MockKernel("test", max_total_local_sizes={0: 64, 1: 1024, 2: 512, 3: 128})

    src = MockDefTemplate(kernels=[kernel])
    program = Program(context.devices, src)

    # The indices here correspond to the devices in the context, not in the platform
    assert program.kernel.test.max_total_local_sizes == {
        context.devices[0]: 1024,
        context.devices[1]: 512,
    }


def test_cannot_override_builtin_globals(mock_context):
    with pytest.raises(
        ValueError, match="'device_params' is a reserved global name and cannot be used"
    ):
        Program(
            [mock_context.device],
            MockDefTemplate(kernels=[MockKernel("test", [None])]),
            render_globals=dict(device_params=None),
        )


def test_builtin_globals(mock_backend_pycuda):
    mock_backend_pycuda.add_devices(
        [PyCUDADeviceInfo(max_threads_per_block=1024), PyCUDADeviceInfo(max_threads_per_block=512)]
    )

    source_template = DefTemplate.from_string(
        "mock_source",
        [],
        """
        KERNEL void test()
        {
            int max_total_local_size = ${device_params.max_total_local_size};
        }
        """,
    )

    api = API.from_api_id(mock_backend_pycuda.api_id)
    context = Context.from_devices([api.platforms[0].devices[0], api.platforms[0].devices[1]])

    src = MockDefTemplate(kernels=[MockKernel("test", [None])], source_template=source_template)

    program = Program(context.devices, src)

    assert "max_total_local_size = 1024" in program.sources[context.devices[0]].source
    assert "max_total_local_size = 512" in program.sources[context.devices[1]].source



from . import dtypes, functions
from .api import (
    API,
    all_api_ids,
    cuda_api_id,
    opencl_api_id,
)
from .array import Array, ArrayLike, MultiArray
from .array_metadata import ArrayMetadataLike
from .buffer import Buffer
from .context import Context
from .device import Device, DeviceFilter
from .device_discovery import (
    platforms_and_devices_by_mask,
    select_devices,
)
from .modules import Module, Snippet
from .platform import Platform, PlatformFilter
from .program import CompilationError, Program
from .queue import MultiQueue, Queue
from .static import StaticKernel
from .template import DefTemplate, RenderError, Template
from .virtual_alloc import VirtualManager
from .vsize import VirtualSizeError


def __getattr__(name: str) -> API:
    if name == "cuda_api":
        return API.from_api_id(cuda_api_id())
    if name == "opencl_api":
        return API.from_api_id(opencl_api_id())
    if name == "any_api":
        apis = API.all_available()
        if len(apis) == 0:
            raise ImportError("No APIs are available. Please install either PyCUDA or PyOpenCL")
        return apis[0]

    raise ImportError(f"Cannot import name '{name}' from '{__name__}'")



"""
Utilities for testing and mocking Grunnur internals.
Brittle, handle with care.
"""

from .mock_base import MockDefTemplate, MockKernel, MockSource
from .mock_factory import MockBackendFactory
from .mock_pycuda import MockPyCUDA, PyCUDADeviceInfo
from .mock_pyopencl import MockPyOpenCL, PyOpenCLDeviceInfo



import pytest

from grunnur import (
    API,
    Context,
    Device,
    DeviceFilter,
    Platform,
    PlatformFilter,
    Queue,
    opencl_api_id,
)
from grunnur.context import BoundMultiDevice


def test_from_devices(mock_backend):
    mock_backend.add_devices(["Device2", "Device3"])

    api = API.from_api_id(mock_backend.api_id)

    platform = api.platforms[0]
    devices = platform.devices[:]
    context = Context.from_devices(devices)
    assert context.platform == platform
    assert [device.as_unbound() for device in context.devices] == devices


def test_from_devices_different_platforms(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2"])
    mock_backend_pyopencl.add_platform_with_devices("Platform2", ["Device3", "Device4"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)

    with pytest.raises(ValueError, match="All devices must belong to the same platform"):
        Context.from_devices([api.platforms[0].devices[0], api.platforms[1].devices[0]])


def test_from_backend_devices_opencl(mock_backend_pyopencl):
    backend = mock_backend_pyopencl

    backend.add_platform_with_devices("Platform1", ["Device1"])
    backend.add_platform_with_devices("Platform2", ["Device2", "Device3"])

    backend_devices = backend.pyopencl.get_platforms()[1].get_devices()
    context = Context.from_backend_devices(backend_devices)

    assert context.platform.name == "Platform2"
    assert [device.name for device in context.devices] == ["Device2", "Device3"]


def test_from_backend_contexts_opencl(mock_backend_pyopencl):
    # OpenCL style - one context, many devices

    backend = mock_backend_pyopencl

    backend.add_platform_with_devices("Platform1", ["Device1"])
    backend.add_platform_with_devices("Platform2", ["Device2", "Device3"])

    backend_devices = backend.pyopencl.get_platforms()[1].get_devices()
    backend_context = backend.pyopencl.Context(backend_devices)

    backend_context2 = backend.pyopencl.Context(backend_devices)
    with pytest.raises(ValueError, match="Cannot make one OpenCL context out of several contexts"):
        Context.from_backend_contexts([backend_context, backend_context2])

    context = Context.from_backend_contexts([backend_context])

    assert context.platform.name == "Platform2"
    assert [device.name for device in context.devices] == ["Device2", "Device3"]

    with pytest.raises(
        TypeError, match="<class 'int'> objects were not recognized as contexts by any API"
    ):
        Context.from_backend_contexts([1])


@pytest.mark.usefixtures("mock_backend_pycuda")
def test_from_backend_contexts_several_apis(mock_backend_pyopencl):
    backend = mock_backend_pyopencl
    backend.add_platform_with_devices("Platform1", ["Device1"])

    backend_devices = backend.pyopencl.get_platforms()[0].get_devices()
    backend_context = backend.pyopencl.Context(backend_devices)

    # Check that when several backends are available,
    # the correct one is used for the given context objects.
    context = Context.from_backend_contexts([backend_context])
    assert context.api.id == opencl_api_id()
    assert context.platform.name == "Platform1"
    assert [device.name for device in context.devices] == ["Device1"]


@pytest.mark.parametrize("take_ownership", [False, True], ids=["no ownership", "take ownership"])
def test_from_backend_contexts_cuda_single_device(mock_backend_pycuda, take_ownership):
    # CUDA style - a context per device

    backend = mock_backend_pycuda

    backend.add_devices(["Device1", "Device2"])

    backend_context = backend.pycuda_driver.Device(1).make_context()

    if not take_ownership:
        # backend context can stay in the stack
        context = Context.from_backend_contexts([backend_context], take_ownership=False)

    else:
        # forgot to pop the backend context off the stack - error
        with pytest.raises(ValueError, match="The given context is already in the context stack"):
            context = Context.from_backend_contexts([backend_context], take_ownership=True)

        backend.pycuda_driver.Context.pop()
        context = Context.from_backend_contexts([backend_context], take_ownership=True)

    # CUDA has no concept of platforms, so the platform name in the mock will be ignored
    assert context.platform.name == "nVidia CUDA"

    assert [device.name for device in context.devices] == ["Device2"]

    if not take_ownership:
        # Clean up - we didn't take ownership of the context, so someone else has to pop it.
        backend.pycuda_driver.Context.pop()


def test_from_backend_contexts_cuda_multi_device(mock_backend_pycuda):
    # CUDA style - a context per device

    backend = mock_backend_pycuda

    backend.add_devices(["Device1", "Device2"])

    backend_context1 = backend.pycuda_driver.Device(0).make_context()
    backend.pycuda_driver.Context.pop()

    backend_context2 = backend.pycuda_driver.Device(1).make_context()
    backend.pycuda_driver.Context.pop()

    # Grunnur mast have ownership
    error_msg = "When dealing with multiple CUDA contexts, Grunnur must be the one managing them"
    with pytest.raises(ValueError, match=error_msg):
        Context.from_backend_contexts([backend_context1, backend_context2])

    context = Context.from_backend_contexts(
        [backend_context1, backend_context2], take_ownership=True
    )

    # CUDA has no concept of platforms, so the platform name in the mock will be ignored
    assert context.platform.name == "nVidia CUDA"

    assert [device.name for device in context.devices] == ["Device1", "Device2"]


def test_from_criteria(mock_backend_pyopencl):
    backend = mock_backend_pyopencl

    backend.add_platform_with_devices("foo-bar", ["Device1"])
    backend.add_platform_with_devices("bar-baz", ["Device2"])
    backend.add_platform_with_devices(
        "foo-baz",
        ["foo-bar", "foo-baz-1", "bar-baz", "foo-baz-1", "foo-baz-2", "foo-baz-2", "foo-baz-3"],
    )

    api = API.from_api_id(backend.api_id)
    context = Context.from_criteria(
        api,
        devices_num=2,
        device_filter=DeviceFilter(
            include_masks=["foo"],
            exclude_masks=["bar"],
            unique_only=True,
        ),
        platform_filter=PlatformFilter(
            include_masks=["foo"],
            exclude_masks=["bar"],
        ),
    )

    assert context.platform.name == "foo-baz"
    assert [device.name for device in context.devices] == ["foo-baz-1", "foo-baz-2"]


def test_bound_device_eq(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)

    platform = api.platforms[0]
    devices = platform.devices[:]
    context = Context.from_devices(devices)

    assert context.devices[0] == context.devices[0]
    assert context.devices[0] != context.devices[1]

    context2 = Context.from_devices(devices)

    assert context2.devices[0] != context.devices[0]


def test_bound_device_str(mock_context):
    s = str(mock_context.device)
    assert mock_context.api.id.shortcut in s
    assert "0,0" in s
    assert "Context" in s


def test_bound_multi_device_creation(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2", "Device3"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)

    platform = api.platforms[0]
    devices = platform.devices[:]
    context = Context.from_devices(devices)
    context2 = Context.from_devices(devices)

    with pytest.raises(
        ValueError, match="All devices in a multi-device must belong to the same context"
    ):
        BoundMultiDevice.from_bound_devices([context.devices[0], context2.devices[1]])

    with pytest.raises(ValueError, match="All devices in a multi-device must be distinct"):
        context.devices[[1, 1]]

    sub_device = context.devices[[2, 1]]
    assert len(sub_device) == 2
    assert sub_device[0].name == "Device3"
    assert sub_device[1].name == "Device2"


def test_bound_multi_device_eq(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2", "Device3"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)
    context = Context.from_devices(api.platforms[0].devices[:])

    assert context.devices[[2, 1]] == context.devices[[2, 1]]
    assert context.devices[[2, 1]] != context.devices[[1, 2]]


def test_bound_multi_device_issubset(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2", "Device3"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)
    context = Context.from_devices(api.platforms[0].devices[:])

    assert context.devices[[2, 1]].issubset(context.devices)


def test_device_shortcut(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2", "Device3"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)
    context = Context.from_devices(api.platforms[0].devices[:])

    with pytest.raises(
        RuntimeError, match="The `device` shortcut only works for single-device contexts"
    ):
        _ = context.device

    context = Context.from_devices([api.platforms[0].devices[2]])
    assert context.device.name == "Device3"


def test_deactivate(mock_backend_pyopencl, mock_backend_pycuda):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1"])
    mock_backend_pycuda.add_devices(["Device1"])

    api = API.from_api_id(mock_backend_pyopencl.api_id)
    context = Context.from_devices([api.platforms[0].devices[0]])
    # Does nothing in OpenCL, but we can still call this for the sake of being generic
    context.deactivate()

    backend_context = mock_backend_pycuda.pycuda_driver.Device(0).make_context()
    backend_context.pop()

    api = API.from_api_id(mock_backend_pycuda.api_id)
    context = Context.from_backend_contexts([backend_context], take_ownership=True)
    assert backend_context.is_stacked()
    context.deactivate()
    assert not backend_context.is_stacked()



import numpy
import pytest

from grunnur import Array, Buffer, MultiArray, MultiQueue, Queue
from grunnur.array_metadata import ArrayMetadata


def _check_array_operations(queue, array_cls):
    arr = numpy.arange(100)
    arr_dev = array_cls.from_host(queue, arr)

    # get to a new array
    assert (arr_dev.get(queue) == arr).all()

    # get to an existing array
    res = numpy.empty(arr.shape, arr.dtype)
    arr_dev.get(queue, dest=res)
    assert (res == arr).all()

    # async get
    res = arr_dev.get(queue, async_=True)
    queue.synchronize()
    assert (res == arr).all()

    # async set
    arr2 = numpy.arange(100) + 1
    arr_dev.set(queue, arr2)
    assert (arr_dev.get(queue) == arr2).all()

    # sync set
    arr2 = numpy.arange(100) + 2
    arr_dev.set(queue, arr2, no_async=True)
    assert (arr_dev.get(queue) == arr2).all()

    # async set from another array
    arr2 = numpy.arange(100) + 3
    arr2_dev = array_cls.from_host(queue, arr2)
    arr_dev.set(queue, arr2_dev)
    assert (arr_dev.get(queue) == arr2).all()

    # sync set from another array
    arr2 = numpy.arange(100) + 4
    arr2_dev = array_cls.from_host(queue, arr2)
    arr_dev.set(queue, arr2_dev, no_async=True)
    assert (arr_dev.get(queue) == arr2).all()


def test_single_device(mock_or_real_context):
    context, _mocked = mock_or_real_context
    _check_array_operations(Queue(context.device), Array)


def test_set_from_non_contiguous(mock_or_real_context):
    context, _mocked = mock_or_real_context

    queue = Queue(context.device)
    arr = Array.empty(context.device, (10, 20), numpy.int32)
    arr2 = Array.empty(context.device, (20, 20), numpy.int32)

    with pytest.raises(
        ValueError, match="Setting from a non-contiguous device array is not supported"
    ):
        arr.set(queue, arr2[::2, :])

    # Can set from a non-contiguous numpy array though
    arr.set(queue, numpy.ones((20, 20), numpy.int32)[::2, :])
    assert (arr.get(queue) == 1).all()


def test_from_host(mock_or_real_context):
    context, _mocked = mock_or_real_context
    queue = Queue(context.device)
    arr = numpy.arange(100)
    arr_dev = Array.from_host(queue, arr)
    assert (arr_dev.get(queue) == arr).all()

    # Create from a host array given only a device - that is, synchronously
    arr_dev = Array.from_host(context.devices[0], arr)
    assert (arr_dev.get(queue) == arr).all()


def test_empty(mock_or_real_context):
    context, _mocked = mock_or_real_context
    queue = Queue(context.device)
    arr_dev = Array.empty(context.device, [100], numpy.int32)
    arr = arr_dev.get(queue)
    assert arr.shape == (100,)
    assert arr.dtype == numpy.int32


def test_empty_like(mock_or_real_context):
    context, _mocked = mock_or_real_context
    arr = numpy.arange(100).reshape(5, 20).astype(numpy.int32)
    arr_dev = Array.empty_like(context.device, arr)
    assert arr_dev.shape == (5, 20)
    assert arr_dev.dtype == numpy.int32


def test_multi_device(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    _check_array_operations(MultiQueue.on_devices([context.devices[0]]), MultiArray)


def test_multi_device_from_host(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    mqueue = MultiQueue.on_devices(context.devices)
    arr = numpy.arange(100)
    arr_dev = MultiArray.from_host(mqueue, arr)
    assert (arr_dev.get(mqueue) == arr).all()


def test_multi_device_empty(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    mqueue = MultiQueue.on_devices(context.devices)

    arr_dev = MultiArray.empty(context.devices, [100], numpy.int32)
    arr = arr_dev.get(mqueue)
    assert arr.shape == (100,)
    assert arr.dtype == numpy.int32

    # explicit device
    arr_dev = MultiArray.empty(context.devices[0:1], [100], numpy.int32)
    assert list(arr_dev.subarrays.keys()) == [context.devices[0]]
    arr = arr_dev.get(mqueue)
    assert arr.shape == (100,)
    assert arr.dtype == numpy.int32

    # explicit splay
    arr_dev = MultiArray.empty(context.devices, [100], numpy.int32, splay=MultiArray.EqualSplay())
    arr = arr_dev.get(mqueue)
    assert arr.shape == (100,)
    assert arr.dtype == numpy.int32

    device0, device1 = context.devices
    assert (arr_dev.subarrays[device0].get(mqueue.queues[device0]) == arr[:50]).all()
    assert (arr_dev.subarrays[device1].get(mqueue.queues[device1]) == arr[50:]).all()


def test_multi_device_mismatched_set(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    mqueue = MultiQueue.on_devices(context.devices)
    arr_dev = MultiArray.empty(context.devices, [100], numpy.int32)
    arr_dev2 = MultiArray.empty(context.devices[0:1], [100], numpy.int32)
    with pytest.raises(
        ValueError, match="Mismatched device sets in the source and the destination"
    ):
        arr_dev.set(mqueue, arr_dev2)


def test_equal_splay(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    mqueue = MultiQueue.on_devices(context.devices)
    arr = numpy.arange(101)
    arr_dev = MultiArray.from_host(mqueue, arr, splay=MultiArray.EqualSplay())

    device0, device1 = context.devices

    assert (arr_dev.subarrays[device0].get(mqueue.queues[device0]) == arr[:51]).all()
    assert (arr_dev.subarrays[device1].get(mqueue.queues[device1]) == arr[51:]).all()

    # Check that the default splay is EqualSplay
    arr_dev = MultiArray.from_host(mqueue, arr)
    assert (arr_dev.subarrays[device0].get(mqueue.queues[device0]) == arr[:51]).all()
    assert (arr_dev.subarrays[device1].get(mqueue.queues[device1]) == arr[51:]).all()

    message = "The number of devices to splay to cannot be greater than the outer array dimension"
    with pytest.raises(ValueError, match=message):
        MultiArray.from_host(
            mqueue, numpy.arange(101).reshape(1, 101), splay=MultiArray.EqualSplay()
        )


def test_clone_splay(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    mqueue = MultiQueue.on_devices(context.devices)
    arr = numpy.arange(101)
    arr_dev = MultiArray.from_host(mqueue, arr, splay=MultiArray.CloneSplay())
    device0, device1 = context.devices
    assert (arr_dev.subarrays[device0].get(mqueue.queues[device0]) == arr).all()
    assert (arr_dev.subarrays[device1].get(mqueue.queues[device1]) == arr).all()


def test_custom_allocator(mock_context):
    context = mock_context
    queue = Queue(context.device)
    allocated = []

    def allocator(device, size):
        allocated.append(size)
        return Buffer.allocate(device, size)

    arr_dev = Array.empty(context.device, [100], numpy.int32, allocator=allocator)
    arr = arr_dev.get(queue)
    assert arr.shape == (100,)
    assert arr.dtype == numpy.int32
    assert allocated == [arr.size * arr.dtype.itemsize]


def test_custom_buffer(mock_context):
    context = mock_context
    queue = Queue(context.device)

    arr = numpy.arange(100).astype(numpy.int32)
    metadata = ArrayMetadata.from_arraylike(arr)

    data = Buffer.allocate(context.device, 100)
    with pytest.raises(
        ValueError, match="Provided data buffer is not big enough to hold the array"
    ):
        Array(metadata, data)

    bigger_data = Buffer.allocate(context.device, arr.size * arr.dtype.itemsize)
    bigger_data.set(queue, arr)

    arr_dev = Array(metadata, bigger_data)
    res = arr_dev.get(queue)
    assert (res == arr).all()


def test_set_checks_shape(mock_context):
    context = mock_context
    queue = Queue(context.device)
    arr = Array.empty(context.device, (10, 20), numpy.int32)

    with pytest.raises(ValueError, match="Shape mismatch: expected \\(10, 20\\), got \\(10, 30\\)"):
        arr.set(queue, numpy.zeros((10, 30), numpy.int32))

    with pytest.raises(ValueError, match="Dtype mismatch: expected int32, got int64"):
        arr.set(queue, numpy.zeros((10, 20), numpy.int64))


def test_set_from_wrong_type(mock_context):
    context = mock_context
    queue = Queue(context.device)
    arr = Array.empty(context.device, (10, 20), numpy.int32)
    with pytest.raises(TypeError, match="Cannot set from an object of type <class 'int'>"):
        arr.set(queue, 1)



import pytest

from grunnur import DeviceFilter, PlatformFilter
from grunnur.testing import PyOpenCLDeviceInfo
from utils import check_select_devices


def test_platform_take_single(mock_stdin, mock_backend_factory, capsys):
    # Test that if only one platform is available, it is selected automatically

    platforms_devices = [("Platform1", ["Device1", "Device2"])]
    inputs = ["1"]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, inputs=inputs, interactive=True
    )

    assert len(devices) == 1
    assert devices[0].name == "Device2"


def test_platform_choose(mock_stdin, mock_backend_factory, capsys):
    # Test selecting one of several available platforms

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4"]),
    ]
    inputs = ["1", "0"]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, inputs=inputs, interactive=True
    )

    assert len(devices) == 1
    assert devices[0].name == "Device3"


def test_platform_take_default(mock_stdin, mock_backend_factory, capsys):
    # Test that just pressing enter gives the default (first) platform

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4"]),
    ]
    inputs = ["", "0"]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, inputs=inputs, interactive=True
    )

    assert len(devices) == 1
    assert devices[0].name == "Device1"


def test_device_take_single(mock_stdin, mock_backend_factory, capsys):
    # Test that if only one device is available, it is selected automatically

    platforms_devices = [("Platform1", ["Device1", "Device2"]), ("Platform2", ["Device3"])]
    inputs = ["1"]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, inputs=inputs, interactive=True
    )

    assert len(devices) == 1
    assert devices[0].name == "Device3"


def test_device_take_multiple(mock_stdin, mock_backend_factory, capsys):
    # Test the case when several devices are requested,
    # and that's exactly how many the chosen platform has.
    # Technically it's the same as a single device and quantity=1, but testing just in case.

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4", "Device5"]),
    ]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, interactive=True, quantity=3
    )

    assert len(devices) == 3
    assert [device.name for device in devices] == ["Device3", "Device4", "Device5"]


def test_device_choose_single(mock_stdin, mock_backend_factory, capsys):
    # Test that if only one device is available, it is selected automatically

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4"]),
    ]
    inputs = ["1", "1"]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, inputs=inputs, interactive=True
    )

    assert len(devices) == 1
    assert devices[0].name == "Device4"


def test_device_choose_multiple(mock_stdin, mock_backend_factory, capsys):
    # Test that if only one device is available, it is selected automatically

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4", "Device5"]),
    ]
    inputs = ["1", "0, 2"]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=inputs,
        interactive=True,
        quantity=2,
    )

    assert len(devices) == 2
    assert [device.name for device in devices] == ["Device3", "Device5"]


def test_device_choose_multiple_wrong_quantity(mock_stdin, mock_backend_factory, capsys):
    # Test error in the case when the user selects a number of devices not equal to `quantity`

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4", "Device5"]),
    ]
    inputs = ["1", "0, 1, 2"]

    with pytest.raises(ValueError, match="Exactly 2 devices must be selected"):
        check_select_devices(
            mock_stdin,
            mock_backend_factory,
            capsys,
            platforms_devices,
            inputs=inputs,
            interactive=True,
            quantity=2,
        )


def test_device_take_default_single(mock_stdin, mock_backend_factory, capsys):
    # Test that just pressing enter gives the default (first) device

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4"]),
    ]
    inputs = ["1", ""]
    devices = check_select_devices(
        mock_stdin, mock_backend_factory, capsys, platforms_devices, inputs=inputs, interactive=True
    )

    assert len(devices) == 1
    assert devices[0].name == "Device3"


def test_device_take_default_multiple(mock_stdin, mock_backend_factory, capsys):
    # Test that just pressing enter gives the first `quantity` devices

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4", "Device5"]),
    ]
    inputs = ["1", ""]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=inputs,
        interactive=True,
        quantity=2,
    )

    assert len(devices) == 2
    assert [device.name for device in devices] == ["Device3", "Device4"]


def test_device_take_all_available(mock_stdin, mock_backend_factory, capsys):
    # Test that `quantity=None` returns all suitable devices.

    platforms_devices = [
        ("Platform1", ["Device1", "Device2"]),
        ("Platform2", ["Device3", "Device4", "Device5"]),
    ]
    inputs = ["1"]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=inputs,
        interactive=True,
        quantity=None,
    )

    assert len(devices) == 3
    assert [device.name for device in devices] == ["Device3", "Device4", "Device5"]


def test_filter_include_platforms(mock_stdin, mock_backend_factory, capsys):
    platforms_devices = [
        ("PlatformFoo", ["Device1", "Device2"]),
        ("PlatformBar", ["Device3", "Device4", "Device5"]),
    ]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        platform_filter=PlatformFilter(include_masks=["Bar"]),
    )

    assert len(devices) == 1
    assert devices[0].name == "Device3"


def test_filter_exclude_platforms(mock_stdin, mock_backend_factory, capsys):
    platforms_devices = [
        ("PlatformFoo", ["Device1", "Device2"]),
        ("PlatformBar", ["Device3", "Device4", "Device5"]),
    ]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        platform_filter=PlatformFilter(exclude_masks=["Foo"]),
    )

    assert len(devices) == 1
    assert devices[0].name == "Device3"


def test_filter_include_devices(mock_stdin, mock_backend_factory, capsys):
    platforms_devices = [
        ("PlatformFoo", ["DeviceFoo", "DeviceBar"]),
        ("PlatformBar", ["DeviceFoo", "DeviceBar", "DeviceBaz"]),
    ]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=["1"],
        interactive=True,
        device_filter=DeviceFilter(include_masks=["Foo"]),
    )

    assert len(devices) == 1
    assert devices[0].name == "DeviceFoo"
    assert devices[0].platform.name == "PlatformBar"


def test_filter_exclude_devices(mock_stdin, mock_backend_factory, capsys):
    platforms_devices = [
        ("PlatformFoo", ["DeviceFoo", "DeviceBar"]),
        ("PlatformBar", ["DeviceFoo", "DeviceBar", "DeviceBaz"]),
    ]
    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=["1", "1"],
        interactive=True,
        device_filter=DeviceFilter(exclude_masks=["Bar"]),
    )

    assert len(devices) == 1
    assert devices[0].name == "DeviceBaz"


def test_filter_exclude_all_devices(mock_stdin, mock_backend_factory, capsys):
    platforms_devices = [
        ("PlatformFoo", ["DeviceFoo", "DeviceBar"]),
        ("PlatformBar", ["DeviceFoo", "DeviceBar", "DeviceBaz"]),
    ]

    with pytest.raises(
        ValueError,
        match="Could not find 1 devices on a single platform matching the given criteria",
    ):
        check_select_devices(
            mock_stdin,
            mock_backend_factory,
            capsys,
            platforms_devices,
            device_filter=DeviceFilter(exclude_masks=["Device"]),
        )


@pytest.mark.parametrize(
    "unique_only", [False, True], ids=["unique_only=False", "unique_only=True"]
)
def test_unique_devices_only(mock_stdin, mock_backend_factory, capsys, unique_only):
    platforms_devices = [
        ("PlatformFoo", ["DeviceFoo", "DeviceFoo"]),
        ("PlatformBar", ["DeviceBar", "DeviceBar", "DeviceBaz"]),
    ]

    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=["1", "1"],
        interactive=True,
        device_filter=DeviceFilter(unique_only=unique_only),
    )

    assert len(devices) == 1
    assert devices[0].name == "DeviceBaz" if unique_only else "DeviceBar"


@pytest.mark.parametrize("exclude_pp", [False, True], ids=["exclude_pp=False", "exclude_pp=True"])
def test_include_pure_parallel_devices(mock_stdin, mock_backend_factory, capsys, exclude_pp):
    # `check_select_devices()` mocks OpenCL, so we can use multiple platforms
    # and OpenCL-specific device info
    platforms_devices = [
        ("PlatformFoo", ["Device1", "Device2"]),
        (
            "PlatformBar",
            [PyOpenCLDeviceInfo(name="Device3", max_work_group_size=1), "Device4", "Device5"],
        ),
    ]

    devices = check_select_devices(
        mock_stdin,
        mock_backend_factory,
        capsys,
        platforms_devices,
        inputs=["1", "1"],
        interactive=True,
        device_filter=DeviceFilter(exclude_pure_parallel=exclude_pp),
    )

    assert len(devices) == 1
    assert devices[0].name == "Device5" if exclude_pp else "Device4"



from __future__ import annotations

import weakref
from abc import ABC
from typing import TYPE_CHECKING

from .adapter_base import QueueAdapter
from .context import BoundDevice, BoundMultiDevice
from .device import Device

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterable, Sequence


class MultiQueue:
    """A queue on multiple devices."""

    devices: BoundMultiDevice
    """Multi-device on which this queue operates."""

    queues: dict[BoundDevice, Queue]
    """Single-device queues associated with device indices."""

    @classmethod
    def on_devices(cls, devices: Iterable[BoundDevice]) -> MultiQueue:
        """Creates a queue from provided devices (belonging to the same context)."""
        return cls([Queue(device) for device in devices])

    def __init__(self, queues: Sequence[Queue]):
        """
        :param queues: single-device queues
            (must belong to distinct devices and the same context).
        """
        self.devices = BoundMultiDevice.from_bound_devices([queue.device for queue in queues])
        self.queues = {queue.device: queue for queue in queues}

    def synchronize(self) -> None:
        """Blocks until queues on all devices are empty."""
        for queue in self.queues.values():
            queue.synchronize()


class Queue:
    """A queue on a single device."""

    device: BoundDevice
    """Device on which this queue operates."""

    def __init__(self, device: BoundDevice):
        """:param device: a device on which to create a queue."""
        self.device = device
        self._queue_adapter = device.context._context_adapter.make_queue_adapter(  # noqa: SLF001
            device._device_adapter  # noqa: SLF001
        )

    def synchronize(self) -> None:
        """Blocks until sub-queues on all devices are empty."""
        self._queue_adapter.synchronize()



# This is a testing module, so it has a lot of assertions.
# ruff: noqa: S101

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

import mako.template

from ..adapter_base import DeviceType
from ..template import DefTemplate

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping, Sequence


class MockKernel:
    def __init__(
        self,
        name: str,
        parameters: Sequence[Any] = (),
        max_total_local_sizes: Mapping[int, tuple[int, ...]] = {},
    ):
        self.name = name
        self.parameters = parameters
        self.max_total_local_sizes = max_total_local_sizes


class MockSource:
    def __init__(
        self,
        *,
        kernels: Sequence[MockKernel] = (),
        prelude: str = "",
        should_fail: bool = False,
        constant_mem: Mapping[str, int] = {},
        source: str | None = None,
        source_template: DefTemplate | None = None,
    ):
        self.name = "mock_source"
        self.kernels = kernels
        self.prelude = prelude
        self.should_fail = should_fail
        self.constant_mem = constant_mem

        # TODO: use two different classes to represent a template source and a template object.
        assert source is None or source_template is None

        self.source = source
        self.source_template = source_template

    def __radd__(self, other: Any) -> MockSource:
        assert isinstance(other, str)
        return MockSource(
            kernels=self.kernels,
            prelude=other + self.prelude,
            should_fail=self.should_fail,
            constant_mem=self.constant_mem,
            source=self.source,
            source_template=self.source_template,
        )

    def split(self, sep: str | None = None) -> list[str]:
        return [*self.prelude.split(sep), "<<< mock source >>>"]

    def __str__(self) -> str:
        return self.name

    def render(self, *args: Any, **kwds: Any) -> MockSource:
        if self.source_template is None:
            return self
        return MockSource(
            kernels=self.kernels,
            prelude=self.prelude,
            should_fail=self.should_fail,
            constant_mem=self.constant_mem,
            source=self.source_template.render(*args, **kwds),
        )


class MockDefTemplate(DefTemplate):
    def __init__(self, *args: Any, **kwds: Any):
        dummy_template = mako.template.Template(text=r"""<%def name="mock_source()"></%def>""")
        super().__init__(
            name="mock_source", mako_def_template=dummy_template.get_def("mock_source"), source=""
        )
        self._mock_source = MockSource(*args, **kwds)

    def render(self, *args: Any, **kwds: Any) -> str:
        # We can't make `MockSource` a subclass of `str` since we need to add attributes to it.
        # So we have to intentionally ignore the return type mismatch.
        return cast(str, self._mock_source.render(*args, **kwds))

    def __str__(self) -> str:
        return str(self._mock_source)



# Arguments are actually used in templates, but Ruff cannot parse it.
# ruff: noqa: ARG005

import pytest

from grunnur.modules import Module, Snippet, render_with_modules
from grunnur.template import DefTemplate, RenderError


def test_snippet_from_callable():
    snippet = Snippet.from_callable(lambda x, y: "${x} + ${y} + ${z}", render_globals=dict(z=3))
    res = render_with_modules("${s(1, 2)}", render_globals=dict(s=snippet)).strip()
    assert res == "1 + 2 + 3"


def test_snippet_from_string():
    snippet = Snippet.from_string("${z}", render_globals=dict(z=3))
    assert render_with_modules(snippet).strip() == "3"

    res = render_with_modules("${s}", render_globals=dict(s=snippet)).strip()
    assert res == "3"


def test_module_from_callable():
    module = Module.from_callable(
        lambda prefix, y: "${prefix} + ${y} + ${z}", render_globals=dict(z=3), name="foo"
    )
    res = render_with_modules("module call: ${m(4)}", render_globals=dict(m=module)).strip()
    # The module's source gets rendered and attached at the beginning of the main template,
    # instead of being rendered inplace like in Snippet's case
    assert res == "_mod_foo_0_ + 4 + 3\n\n\nmodule call: _mod_foo_0_"


def test_module_from_string():
    module = Module.from_string("${prefix} + ${z}", render_globals=dict(z=3), name="foo")
    res = render_with_modules("module call: ${m}", render_globals=dict(m=module)).strip()
    # The module's source gets rendered and attached at the beginning of the main template,
    # instead of being rendered inplace like in Snippet's case
    assert res == "_mod_foo_0_ + 3\n\n\nmodule call: _mod_foo_0_"


def test_render_snippet():
    snippet = Snippet.from_callable(lambda x, y: "${x} + ${y} + ${z}", render_globals=dict(z=3))
    assert render_with_modules(snippet, render_args=[1, 2]).strip() == "1 + 2 + 3"


def test_render_snippet_with_render_globals():
    # Check that provided render globals are added to those of the snippet
    snippet = Snippet.from_callable(
        lambda x, y: "${x} + ${y} + ${z} + ${q}", render_globals=dict(z=3)
    )
    assert (
        render_with_modules(snippet, render_args=[1, 2], render_globals=dict(q=4)).strip()
        == "1 + 2 + 3 + 4"
    )
    with pytest.raises(ValueError, match="Cannot add a global 'z' - it already exists"):
        render_with_modules(snippet, render_args=[1, 2], render_globals=dict(z=5, q=4))


def test_render_string():
    assert render_with_modules("abcde").strip() == "abcde"


def test_render_string_with_args():
    with pytest.raises(ValueError, match="A textual source cannot have `render_args` set."):
        render_with_modules("abcde", render_args=[1, 2])


def test_render_callable():
    res = render_with_modules(
        lambda x, y: "${x} + ${y} + ${z}", render_args=[1, 2], render_globals=dict(z=3)
    ).strip()
    assert res == "1 + 2 + 3"


def test_render_def_template():
    tmpl = DefTemplate.from_callable("test", lambda x, y: "${x} + ${y} + ${z}")
    res = render_with_modules(tmpl, render_args=[1, 2], render_globals=dict(z=3)).strip()
    assert res == "1 + 2 + 3"


def test_render_unknown_type():
    with pytest.raises(TypeError):
        render_with_modules(1)


def test_render_error():
    module = Module.from_callable(
        lambda prefix, x: "${prefix} + ${x} + ${bar}", render_globals=dict(baz=3), name="foo"
    )
    with pytest.raises(RenderError) as exc_info:
        render_with_modules("module call: ${m(1)}", render_globals=dict(m=module))

    assert exc_info.type == RenderError

    # Check that we get a correct info from a render error nested in the module hierarchy
    e = exc_info.value

    assert e.args == ("_mod_foo_0_", 1)
    assert e.globals == dict(baz=3)
    assert type(e.exception) is NameError
    assert e.source == module.template.source


class CustomObj:
    """A class supporting custom module processing."""

    def __init__(self, module):
        self.module = module

    def __process_modules__(self, process):
        return RenderableCustomObj(process(self.module))


class RenderableCustomObj:
    def __init__(self, processed_module):
        self.module = processed_module


def test_process_objects():
    # Checks that all supported types of objects are correctly traversed
    # in search for Modules.

    m1 = Module.from_string("m1: ${prefix}", name="m1")
    m2 = Module.from_string("m2: ${prefix}", name="m2")
    m3 = Module.from_string("m3: ${prefix}", name="m3")
    m4 = Module.from_string("m4: ${prefix}", name="m4")
    m5 = Module.from_string("m5: ${prefix}", name="m5")

    res = render_with_modules(
        """
        ${module_obj}
        ${type(custom_obj) == ref_type} ${custom_obj.module}
        ${dict_obj['module']}
        ${list_obj[0]}
        ${tuple_obj[0]}
        ${non_module_obj}
        """,
        render_globals=dict(
            module_obj=m1,
            custom_obj=CustomObj(m2),
            ref_type=RenderableCustomObj,
            dict_obj=dict(module=m3),
            list_obj=[m4],
            tuple_obj=(m5,),
            non_module_obj=1,
        ),
    ).strip()

    assert res == (
        "m1: _mod_m1_0_\n\n\n"
        "m2: _mod_m2_1_\n\n\n"
        "m3: _mod_m3_2_\n\n\n"
        "m4: _mod_m4_3_\n\n\n"
        "m5: _mod_m5_4_\n\n\n\n"
        "        _mod_m1_0_\n"
        "        True _mod_m2_1_\n"
        "        _mod_m3_2_\n"
        "        _mod_m4_3_\n"
        "        _mod_m5_4_\n"
        "        1"
    )


def test_module_cache():
    # a module with no parameters
    m1 = Module.from_callable(lambda prefix: "m1: ${prefix}", name="m1")

    # a module with several parameters
    m2 = Module.from_callable(lambda prefix, x, y: "m2: ${prefix} ${x} ${y}", name="m2")

    res = render_with_modules(
        """
        ${m1}
        ${m1}
        ${m2(1, 2)}
        ${m2(2, 3)}
        ${m2(1, 2)}
        """,
        render_globals=dict(m1=m1, m2=m2),
    ).strip()

    assert res == (
        "m1: _mod_m1_0_\n\n\n"
        "m2: _mod_m2_1_ 1 2\n\n\n"
        "m2: _mod_m2_2_ 2 3\n\n\n\n"
        "        _mod_m1_0_\n"
        "        _mod_m1_0_\n"
        "        _mod_m2_1_\n"
        "        _mod_m2_2_\n"
        "        _mod_m2_1_"
    )



import pytest

from grunnur.sorted_list import SortedList


def test_basics():
    s = SortedList([2, 4, 6, 1, 3, 9], key=lambda x: x)
    assert len(s) == 6
    assert s[3] == 4
    assert list(s) == [1, 2, 3, 4, 6, 9]


def test_index():
    s = SortedList([2, 4, 6, 1, 3, 9], key=lambda x: x)
    assert s.index(4) == 3
    with pytest.raises(ValueError, match="0 is not in list"):
        s.index(0)


def test_insert():
    s = SortedList([2, 4, 6, 1, 3, 9], key=lambda x: x)
    s.insert(5)
    assert list(s) == [1, 2, 3, 4, 5, 6, 9]

    # Check that if there are several elements with the same key,
    # the new item is inserted on the left.
    s = SortedList([2, 4, 6, 1, 5, 9], key=lambda x: x // 2)
    s.insert(5)
    assert list(s) == [1, 2, 5, 4, 5, 6, 9]


def test_remove():
    s = SortedList([2, 4, 6, 1, 3, 9], key=lambda x: x)
    s.remove(3)
    assert list(s) == [1, 2, 4, 6, 9]


def test_argfind_ge():
    s = SortedList([2, 4, 6, 1, 3, 9], key=lambda x: x)
    assert s.argfind_ge(8) == 5
    with pytest.raises(ValueError, match="No item found with key at or above: 10"):
        s.argfind_ge(10)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, overload

from .template import DefTemplate, RenderError
from .utils import update_dict

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Mapping, Sequence


class Snippet:
    """
    Contains a source snippet - a template function that will be rendered in place,
    with possible context that can include other :py:class:`Snippet`
    or :py:class:`Module` objects.
    """

    def __init__(self, template: DefTemplate, render_globals: Mapping[str, Any] = {}):
        """Creates a snippet out of a prepared template."""
        self.name = template.name
        self.template = template
        self.render_globals = render_globals

    def with_added_globals(self, add_globals: Mapping[str, Any] = {}) -> Snippet:
        new_globals = update_dict(
            self.render_globals,
            add_globals,
            error_msg="Cannot add a global '{name}' - it already exists",
        )
        return Snippet(self.template, new_globals)

    @classmethod
    def from_callable(
        cls,
        callable_obj: Callable[..., str],
        name: str = "_snippet",
        render_globals: Mapping[str, Any] = {},
    ) -> Snippet:
        """
        Creates a snippet from a callable returning a string.
        The parameter list of the callable is used to create the pararameter list
        of the resulting template def; the callable should return the body of a
        Mako template def regardless of the arguments it receives.

        :param callable_obj: a callable returning the template source.
        :param name: the snippet's name (will simplify debugging)
        :param render_globals: a dictionary of "globals" to be used when rendering the template.
        """
        template = DefTemplate.from_callable(name, callable_obj)
        return cls(template, render_globals=render_globals)

    @classmethod
    def from_string(
        cls, source: str, name: str = "_snippet", render_globals: Mapping[str, Any] = {}
    ) -> Snippet:
        """
        Creates a snippet from a template source, treated as a body of a
        template def with no arguments.

        :param source: a string with the template source.
        :param name: the snippet's name (will simplify debugging)
        :param render_globals: a dictionary of "globals" to be used when rendering the template.
        """
        template = DefTemplate.from_string(name, [], source)
        return cls(template, render_globals=render_globals)

    def __process_modules__(
        self, process: Callable[[Mapping[str, Any]], Mapping[str, Any]]
    ) -> RenderableSnippet:
        return RenderableSnippet(self.template, process(self.render_globals))


class RenderableSnippet:
    """A snippet with processed dependencies and ready to be rendered."""

    def __init__(self, template: DefTemplate, render_globals: Mapping[str, Any]):
        self.template = template
        self.render_globals = render_globals

    def __call__(self, *args: Any) -> str:
        return self.template.render(*args, **self.render_globals)

    def __str__(self) -> str:
        return self()


class Module:
    """
    Contains a source module - a template function that will be rendered at root level,
    and the place where it was called will receive its unique identifier (prefix),
    which is used to prefix all module's functions, types and macros in the global namespace.
    """

    @classmethod
    def from_callable(
        cls,
        callable_obj: Callable[..., str],
        name: str = "_module",
        render_globals: Mapping[str, Any] = {},
    ) -> Module:
        """
        Creates a module from a callable returning a string.
        The parameter list of the callable is used to create the pararameter list
        of the resulting template def; the callable should return the body of a
        Mako template def regardless of the arguments it receives.

        The prefix will be passed as the first argument to the template def on render.

        :param callable_obj: a callable returning the template source.
        :param name: the module's name (will simplify debugging)
        :param render_globals: a dictionary of "globals" to be used when rendering the template.
        """
        template = DefTemplate.from_callable(name, callable_obj)
        return cls(template, render_globals=render_globals)

    @classmethod
    def from_string(
        cls, source: str, name: str = "_module", render_globals: Mapping[str, Any] = {}
    ) -> Module:
        """
        Creates a module from a template source, treated as a body of a
        template def with a single argument (prefix).

        :param source: a string with the template source.
        :param name: the module's name (will simplify debugging)
        :param render_globals: a dictionary of "globals" to be used when rendering the template.
        """
        template = DefTemplate.from_string(name, ["prefix"], source)
        return cls(template, render_globals=render_globals)

    def __init__(self, template: DefTemplate, render_globals: Mapping[str, Any] = {}):
        """
        Creates a module out of a prepared template.

        :param template:
        :param render_globals:
        """
        self.name = template.name
        self.template = template
        self.render_globals = render_globals

    def process(self, collector: SourceCollector) -> RenderableModule:
        return RenderableModule(
            collector, id(self), self.template, process(self.render_globals, collector)
        )


class RenderableModule:
    """A module with processed dependencies and ready to be rendered."""

    def __init__(
        self,
        collector: SourceCollector,
        module_id: int,
        template: DefTemplate,
        render_globals: Mapping[str, Any],
    ):
        self.module_id = module_id
        self.collector = collector
        self.template = template
        self.render_globals = render_globals

    def __call__(self, *args: Any) -> str:
        return self.collector.add_module(self.module_id, self.template, args, self.render_globals)

    def __str__(self) -> str:
        return self()


class SourceCollector:
    def __init__(self) -> None:
        self.module_cache: dict[tuple[int, tuple[Any, ...]], str] = {}
        self.sources: list[str] = []
        self.prefix_counter = 0

    def add_module(
        self,
        module_id: int,
        template: DefTemplate,
        args: Sequence[Any],
        render_globals: Mapping[str, Any],
    ) -> str:
        # This caching serves two purposes.
        # First, it reduces the amount of generated code by not generating
        # the same module several times.
        # Second, if the same module object is used in other modules,
        # the data structures defined in this module will be suitable
        # for functions in these modules.
        call_id = (module_id, tuple(args))
        if call_id in self.module_cache:
            return self.module_cache[call_id]

        prefix = "_mod_" + template.name + "_" + str(self.prefix_counter) + "_"
        self.prefix_counter += 1

        src = template.render(prefix, *args, **render_globals)
        self.sources.append(src)

        self.module_cache[call_id] = prefix

        return prefix

    def get_source(self) -> str:
        return "\n".join(self.sources)


@overload
def process(obj: Module, collector: SourceCollector) -> RenderableModule: ...


@overload
def process(obj: Snippet, collector: SourceCollector) -> RenderableSnippet: ...


@overload
def process(obj: Mapping[str, Any], collector: SourceCollector) -> dict[str, Any]: ...


@overload
def process(obj: list[Any], collector: SourceCollector) -> list[Any]: ...


@overload
def process(obj: tuple[Any, ...], collector: SourceCollector) -> tuple[Any, ...]: ...


def process(obj: Any, collector: SourceCollector) -> Any:
    if isinstance(obj, Module):
        return obj.process(collector)
    if hasattr(obj, "__process_modules__"):
        return obj.__process_modules__(lambda x: process(x, collector))
    # Since we can't guarantee that custom types derived from these ones
    # will have the same constructors, we have to do strict type checking.
    if type(obj) == dict:  # noqa: E721
        return {k: process(v, collector) for k, v in obj.items()}
    if type(obj) == tuple:  # noqa: E721
        return tuple(process(v, collector) for v in obj)
    if type(obj) == list:  # noqa: E721
        return [process(v, collector) for v in obj]
    return obj


def render_with_modules(
    src: str | Callable[..., str] | DefTemplate | Snippet,
    render_args: Sequence[Any] = (),
    render_globals: Mapping[str, Any] = {},
) -> str:
    """
    Renders the given source traversing the arguments and globals processing modules.
    If a module is attempted to be rendered, its source is prepended to the resulting source,
    and the caller receives the generated module prefix.

    Nested arguments/globals of :py:class:`Module` and :py:class:`Snippet`
    are traversed automatically. Also traversed are instances of `dict`, `list` and `tuple`.
    Other classes must define a `__process_modules__(self, process)` method,
    where `process` is a one-argument function traversing any of the supported objects
    and returning the resulting object with all nested :py:class:`Module` objects
    changed to :py:class:`RenderableModule`.

    If ``src`` is a string, a callable or a :py:class:`DefTemplate`,
    a :py:class:`Snippet` is created with a corresponding classmethod or the constructor.

    If ``src`` is a :py:class:`Snippet`, ``render_globals`` will be added to its render globals
    (a ``ValueError`` will be thrown if there is a name clash).

    If any of the nested templates fails to render,
    a :py:class:`~grunnur.template.RenderError` is propagated
    from that place to this function, and raised here.

    :param src: the textual source, template or a snippet to render.
    :param render_args: a list of arguments to pass to the template def.
    :param render_globals: a dict of globals to render the template with.
    """
    collector = SourceCollector()
    render_args = process(tuple(render_args), collector)

    if isinstance(src, str):
        if len(render_args) > 0:
            raise ValueError("A textual source cannot have `render_args` set.")
        snippet = Snippet.from_string(src, name="_main_", render_globals=render_globals)
    elif isinstance(src, DefTemplate):
        snippet = Snippet(src, render_globals=render_globals)
    elif isinstance(src, Snippet):
        snippet = src.with_added_globals(render_globals)
    elif callable(src):
        snippet = Snippet.from_callable(src, name="_main_", render_globals=render_globals)
    else:
        raise TypeError(f"Cannot render an object of type {type(src)}")

    main_renderable = process(snippet, collector)

    try:
        main_src = main_renderable(*render_args)
    except RenderError as e:
        # The error will come from a chain of modules and snippets rendering each other,
        # so it will be buried deep in the traceback.
        # Setting the cause to None to cut all the intermediate calls which don't carry
        # any important information.
        raise e from None

    return collector.get_source() + "\n" + main_src



from __future__ import annotations

import itertools
from collections import Counter
from math import ceil, floor, sqrt
from typing import TYPE_CHECKING, Any, NamedTuple

from .modules import Module, RenderableModule, Snippet
from .template import Template
from .utils import min_blocks, prod

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Iterator, Mapping, Sequence


TEMPLATE = Template.from_associated_file(__file__)


def _factorize_step(x: int) -> int:
    return 1 + (x << 2) - ((x >> 1) << 1)


def factorize(num: int) -> list[int]:
    """Factorizes the given integer."""
    maxq = int(floor(sqrt(num)))
    d = 1
    q = 2 if num % 2 == 0 else 3
    while q <= maxq and num % q != 0:
        q = _factorize_step(d)
        d += 1
    return [q, *factorize(num // q)] if q <= maxq else [num]


class PrimeFactors:
    """Contains a natural number's decomposition into prime factors."""

    def __init__(self, factors: Mapping[int, int]):
        self.factors = factors

    @classmethod
    def decompose(cls, num: int) -> PrimeFactors:
        factors_list = factorize(num)
        factors = Counter(factors_list)
        return cls(dict(factors))

    def get_value(self) -> int:
        res = 1
        for pwr, exp in self.factors.items():
            res *= pwr**exp
        return res

    def get_arrays(self) -> tuple[list[int], list[int]]:
        bases = list(self.factors.keys())
        exponents = [self.factors[base] for base in bases]
        return bases, exponents

    def div_by(self, other: PrimeFactors) -> PrimeFactors:
        # assumes that `self` is a multiple of `other`
        factors = dict(self.factors)
        for o_pwr, o_exp in other.factors.items():
            factors[o_pwr] -= o_exp
            if factors[o_pwr] == 0:
                del factors[o_pwr]
        return PrimeFactors(factors)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, PrimeFactors) and self.factors == other.factors


def _get_decompositions(num_factors: PrimeFactors, parts: int) -> Iterator[list[int]]:
    """
    Helper recursive function for ``get_decompositions()``.
    Iterates over all possible decompositions of ``num_factors`` into ``parts`` factors.
    """
    if parts == 1:
        yield [num_factors.get_value()]
        return

    bases, exponents = num_factors.get_arrays()
    for sub_exps in itertools.product(*[range(exp, -1, -1) for exp in exponents]):
        part_factors = PrimeFactors(
            {pwr: sub_exp for pwr, sub_exp in zip(bases, sub_exps, strict=True) if sub_exp > 0}
        )
        part = part_factors.get_value()
        remainder = num_factors.div_by(part_factors)
        for decomp in _get_decompositions(remainder, parts - 1):
            yield [part, *decomp]


def get_decompositions(num: int, parts: int) -> Iterator[list[int]]:
    """Iterates over all possible decompositions of ``num`` into ``parts`` factors."""
    num_factors = PrimeFactors.decompose(num)
    return _get_decompositions(num_factors, parts)


def find_local_size_decomposition(
    global_size: Sequence[int], flat_local_size: int, threshold: float = 0.05
) -> list[int]:
    """
    Returns a tuple of the same size as ``global_size``,
    with the product equal to ``flat_local_size``,
    and minimal difference between ``product(global_size)``
    and ``product(min_blocks(gs, ls) * ls for gs, ls in zip(global_size, local_size))``
    (i.e. tries to minimize the amount of empty threads).
    """
    flat_global_size = prod(global_size)

    if flat_local_size > flat_global_size:
        raise ValueError("The given local size does not fit in the given global size")

    if flat_local_size == flat_global_size:
        return list(global_size)

    threads_num = prod(global_size)

    best_ratio: float | None = None
    best_local_size: list[int] | None = None

    for local_size in get_decompositions(flat_local_size, len(global_size)):
        bounding_global_size = [
            ls * min_blocks(gs, ls) for gs, ls in zip(global_size, local_size, strict=True)
        ]
        empty_threads = prod(bounding_global_size) - threads_num
        ratio = empty_threads / threads_num

        # Stopping iteration early, because there may be a lot of elements to iterate over,
        # and we do not need the perfect solution.
        if ratio < threshold:
            return local_size

        if best_ratio is None or ratio < best_ratio:
            best_ratio = ratio
            best_local_size = local_size

    # If flat_local_size <= product(global_size),
    # there is at least one suitable decomposition (flat_local_size, 1, 1, ...).
    assert best_local_size is not None  # noqa: S101

    return best_local_size


def _group_dimensions(
    vdim: int, virtual_shape: Sequence[int], adim: int, available_shape: Sequence[int]
) -> tuple[list[list[int]], list[list[int]]]:
    """
    ``vdim`` and ``adim`` are used for the absolute addressing
    of dimensions during recursive calls.
    """
    if len(virtual_shape) == 0:
        return [], []

    vdim_group = 1  # number of currently grouped virtual dimensions
    adim_group = 1  # number of currently grouped available dimensions

    while True:
        # If we have more elements in the virtual group than there is in the available group,
        # extend the available group by one dimension.
        if prod(virtual_shape[:vdim_group]) > prod(available_shape[:adim_group]):
            adim_group += 1
            continue

        # If the remaining available dimensions cannot accommodate the remaining virtual dimensions,
        # we try to fit one more virtual dimension in the virtual group.
        if prod(virtual_shape[vdim_group:]) > prod(available_shape[adim_group:]):
            vdim_group += 1
            continue

        # If we are here, it means that:
        # 1) the current available group can accommodate the current virtual group;
        # 2) the remaining available dimensions can accommodate the remaining virtual dimensions.
        # This means we can make a recursive call now.

        # Attach any following trivial virtual dimensions (of size 1) to this group
        # This will help to avoid unassigned trivial dimensions with no real dimensions left.
        while vdim_group < len(virtual_shape) and virtual_shape[vdim_group] == 1:
            vdim_group += 1

        v_res = list(range(vdim, vdim + vdim_group))
        a_res = list(range(adim, adim + adim_group))

        v_remainder, a_remainder = _group_dimensions(
            vdim + vdim_group,
            virtual_shape[vdim_group:],
            adim + adim_group,
            available_shape[adim_group:],
        )
        return [v_res, *v_remainder], [a_res, *a_remainder]


def group_dimensions(
    virtual_shape: Sequence[int], available_shape: Sequence[int]
) -> tuple[list[list[int]], list[list[int]]]:
    """
    Determines which available dimensions the virtual dimensions can be embedded into.
    Prefers using the maximum number of available dimensions, since in that case
    less divisions will be needed to calculate virtual indices based on the real ones.

    Returns two lists, one of tuples with indices of grouped virtual dimensions, the other
    one of tuples with indices of corresponding group of available dimensions,
    such that for any group of virtual dimensions, the total number of elements they cover
    does not exceed the number of elements covered by the
    corresponding group of available dimensions.

    Dimensions are grouped in order, so tuples in both lists, if concatenated,
    give `(0 ... len(virtual_shape)-1)` and `(0 ... n)`, where `n < len(available_shape`.
    """
    # Already checked in `VirtualSizes` constructor
    assert prod(virtual_shape) <= prod(available_shape)  # noqa: S101
    return _group_dimensions(0, virtual_shape, 0, available_shape)


def find_bounding_shape(virtual_size: int, available_shape: Sequence[int]) -> list[int]:
    """
    Finds a tuple of the same length as ``available_shape``, with every element
    not greater than the corresponding element of ``available_shape``,
    and product not lower than ``virtual_size`` (trying to minimize that difference).
    """
    # TODO: in most cases it is possible to find such a tuple that `prod(result) == virtual_size`,
    # but the current algorithm does not gurantee it. Finding such a tuple would
    # eliminate some empty threads.

    # Already checked in `VirtualSizes` constructor
    assert virtual_size <= prod(available_shape)  # noqa: S101

    free_size = virtual_size
    free_dims = set(range(len(available_shape)))
    bounding_shape = [0] * len(available_shape)

    # The loop terminates, since `virtual_size` is guaranteed to fit into `available_shape`
    # (worst case scenario, the result will be just `available_shape` itself).
    while True:
        guess = ceil(free_size ** (1 / len(free_dims)))
        fixed_size = 1
        for fdim in list(free_dims):
            if guess > available_shape[fdim]:
                fixed_size *= available_shape[fdim]
                free_dims.remove(fdim)
                bounding_shape[fdim] = available_shape[fdim]
            else:
                bounding_shape[fdim] = guess

        if fixed_size == 1:
            break

        free_size = min_blocks(free_size, fixed_size)

    return bounding_shape


class ShapeGroups:
    def __init__(self, virtual_shape: Sequence[int], available_shape: Sequence[int]):
        # A mapping from a dimension in the virtual shape to a tuple of dimensions
        # in the real shape it uses (and possibly shares with other virtual dimensions).
        self.real_dims: dict[int, list[int]] = {}

        # A mapping from a dimension in the virtual shape to a tuple of strides
        # used to get a flat index in the group of real dimensions it uses.
        self.real_strides: dict[int, list[int]] = {}

        # A mapping from a dimension in the virtual shape to the stride that is used to extract it
        # from the flat index obtained from the corresponding group of real dimensions.
        self.virtual_strides: dict[int, int] = {}

        # A mapping from a dimension in the virtual shape to the major dimension
        # (the one with the largest stride) in the group of virtual dimensions it belongs to
        # (the group includes all virtual dimensions using a certain subset of real dimensions).
        self.major_vdims: dict[int, int] = {}

        # The actual shape used to enqueue the kernel.
        self.bounding_shape: list[int] = []

        # A list of tuples `(threshold, stride_info)` used for skipping unused threads.
        # `stride_info` is a list of 2-tuples `(real_dim, stride)` used to construct
        # a flat index from several real dimensions, and then compare it with the threshold.
        self.skip_thresholds: list[tuple[int, list[tuple[int, int]]]] = []

        v_groups, a_groups = group_dimensions(virtual_shape, available_shape)

        for v_group, a_group in zip(v_groups, a_groups, strict=True):
            virtual_subshape = virtual_shape[v_group[0] : v_group[-1] + 1]
            virtual_subsize = prod(virtual_subshape)

            bounding_subshape = find_bounding_shape(
                virtual_subsize, available_shape[a_group[0] : a_group[-1] + 1]
            )

            self.bounding_shape += bounding_subshape

            if virtual_subsize < prod(bounding_subshape):
                strides = [(adim, prod(bounding_subshape[:i])) for i, adim in enumerate(a_group)]
                self.skip_thresholds.append((virtual_subsize, strides))

            for vdim in v_group:
                self.real_dims[vdim] = a_group
                self.real_strides[vdim] = [
                    prod(self.bounding_shape[a_group[0] : adim]) for adim in a_group
                ]
                self.virtual_strides[vdim] = prod(virtual_shape[v_group[0] : vdim])

                # The major virtual dimension (the one that does not require
                # modulus operation when extracting its index from the flat index)
                # is the last non-trivial one (not of size 1).
                # Modulus will not be optimized away by the compiler,
                # but we know that all threads outside of the virtual group will be
                # filtered out by VIRTUAL_SKIP_THREADS.
                for major_vdim in range(len(v_group) - 1, -1, -1):
                    if virtual_shape[v_group[major_vdim]] > 1:
                        break

                self.major_vdims[vdim] = v_group[major_vdim]


class VsizeModules(NamedTuple):
    """
    A collection of modules passed to :py:class:`grunnur.StaticKernel`.
    Should be used instead of regular group/thread id functions.
    """

    local_id: Module
    """
    Provides the function ``VSIZE_T ${local_id}(int dim)``
    returning the local id of the current thread.
    """

    local_size: Module
    """
    Provides the function ``VSIZE_T ${local_size}(int dim)``
    returning the size of the current group.
    """

    group_id: Module
    """
    Provides the function ``VSIZE_T ${group_id}(int dim)``
    returning the group id of the current thread.
    """

    num_groups: Module
    """
    Provides the function ``VSIZE_T ${num_groups}(int dim)``
    returning the number of groups in dimension  ``dim``.
    """

    global_id: Module
    """
    Provides the function ``VSIZE_T ${global_id}(int dim)``
    returning the global id of the current thread.
    """

    global_size: Module
    """
    Provides the function ``VSIZE_T ${global_size}(int dim)``
    returning the global size along dimension ``dim``."""

    global_flat_id: Module
    """
    Provides the function ``VSIZE_T ${global_flat_id}()``
    returning the global id of the current thread with all dimensions flattened.
    """

    global_flat_size: Module
    """
    Provides the function ``VSIZE_T ${global_flat_size}()``.
    returning the global size of with all dimensions flattened.
    """

    skip: Module
    """
    Provides the function ``bool ${skip}()`` that should be used
    at the start of a static kernel function to see if the current thread/work item
    is inside the padding area and needs to be skipped.
    Usually one would write ``if (${skip}()) return;``.
    """

    def __process_modules__(
        self, process: Callable[[Module], RenderableModule]
    ) -> VsizeRenderableModules:
        return VsizeRenderableModules(
            local_id=process(self.local_id),
            local_size=process(self.local_size),
            group_id=process(self.group_id),
            num_groups=process(self.num_groups),
            global_id=process(self.global_id),
            global_size=process(self.global_size),
            global_flat_id=process(self.global_flat_id),
            global_flat_size=process(self.global_flat_size),
            skip=process(self.skip),
        )

    @classmethod
    def from_shape_data(
        cls,
        virtual_global_size: Sequence[int],
        virtual_local_size: Sequence[int],
        bounding_global_size: Sequence[int],
        virtual_grid_size: Sequence[int],
        local_groups: ShapeGroups,
        grid_groups: ShapeGroups,
    ) -> VsizeModules:
        local_id_mod = Module(
            TEMPLATE.get_def("local_id"),
            render_globals=dict(virtual_local_size=virtual_local_size, local_groups=local_groups),
        )

        local_size_mod = Module(
            TEMPLATE.get_def("local_size"),
            render_globals=dict(virtual_local_size=virtual_local_size),
        )

        group_id_mod = Module(
            TEMPLATE.get_def("group_id"),
            render_globals=dict(virtual_grid_size=virtual_grid_size, grid_groups=grid_groups),
        )

        num_groups_mod = Module(
            TEMPLATE.get_def("num_groups"), render_globals=dict(virtual_grid_size=virtual_grid_size)
        )

        global_id_mod = Module(
            TEMPLATE.get_def("global_id"),
            render_globals=dict(
                local_id_mod=local_id_mod, group_id_mod=group_id_mod, local_size_mod=local_size_mod
            ),
        )

        global_size_mod = Module(
            TEMPLATE.get_def("global_size"),
            render_globals=dict(virtual_global_size=virtual_global_size),
        )

        global_flat_id_mod = Module(
            TEMPLATE.get_def("global_flat_id"),
            render_globals=dict(
                virtual_global_size=virtual_global_size, global_id_mod=global_id_mod, prod=prod
            ),
        )

        global_flat_size_mod = Module(
            TEMPLATE.get_def("global_flat_size"),
            render_globals=dict(
                global_size_mod=global_size_mod, virtual_global_size=virtual_global_size
            ),
        )

        skip_local_threads_mod = Module(
            TEMPLATE.get_def("skip_local_threads"), render_globals=dict(local_groups=local_groups)
        )

        skip_groups_mod = Module(
            TEMPLATE.get_def("skip_groups"), render_globals=dict(grid_groups=grid_groups)
        )

        skip_global_threads_mod = Module(
            TEMPLATE.get_def("skip_global_threads"),
            render_globals=dict(
                virtual_global_size=virtual_global_size,
                bounding_global_size=bounding_global_size,
                global_id_mod=global_id_mod,
            ),
        )

        skip = Module(
            TEMPLATE.get_def("skip"),
            render_globals=dict(
                skip_local_threads_mod=skip_local_threads_mod,
                skip_groups_mod=skip_groups_mod,
                skip_global_threads_mod=skip_global_threads_mod,
            ),
        )

        return cls(
            local_id=local_id_mod,
            local_size=local_size_mod,
            group_id=group_id_mod,
            num_groups=num_groups_mod,
            global_id=global_id_mod,
            global_size=global_size_mod,
            global_flat_id=global_flat_id_mod,
            global_flat_size=global_flat_size_mod,
            skip=skip,
        )


class VsizeRenderableModules(NamedTuple):
    local_id: RenderableModule
    local_size: RenderableModule
    group_id: RenderableModule
    num_groups: RenderableModule
    global_id: RenderableModule
    global_size: RenderableModule
    global_flat_id: RenderableModule
    global_flat_size: RenderableModule
    skip: RenderableModule


class VirtualSizeError(Exception):
    """Raised when a virtual size cannot be found due to device limitations."""


class VirtualSizes:
    real_local_size: tuple[int, ...]
    real_global_size: tuple[int, ...]
    vsize_modules: VsizeModules

    def __init__(
        self,
        max_total_local_size: int,
        max_local_sizes: Sequence[int],
        max_num_groups: Sequence[int],
        local_size_multiple: int,
        virtual_global_size: Sequence[int],
        virtual_local_size: Sequence[int] | None = None,
    ):
        if virtual_local_size is not None and len(virtual_local_size) != len(virtual_global_size):
            raise ValueError("Global size and local size must have the same number of dimensions")

        # Since the device uses column-major ordering of sizes, while we get
        # row-major ordered shapes, we invert our shapes
        # to facilitate internal handling.
        virtual_global_size = list(reversed(virtual_global_size))
        if virtual_local_size is not None:
            virtual_local_size = list(reversed(virtual_local_size))

        # In device parameters `max_total_local_size` is >= any of `max_local_sizes`,
        # but it can be overridden to get a kernel that uses less resources.
        max_local_sizes = [min(max_total_local_size, mls) for mls in max_local_sizes]

        # Defensive adjustment #1.
        # With consistent device parameters, maximum flat local size won't be greater
        # than the product of maximum local sizes for each dimension.
        # But if it is, we just limit it to that value.
        max_total_local_size = min(max_total_local_size, prod(max_local_sizes))

        # Defensive adjustment #2.
        # It would be very strange if a device had a local size multiple
        # so big you can't actually launch that many threads.
        # But if it does, we limit it.
        local_size_multiple = min(local_size_multiple, max_total_local_size)

        if virtual_local_size is None:
            # TODO: We can obtain better results by taking occupancy into account here,
            # but for now we will assume that the more threads, the better.
            flat_global_size = prod(virtual_global_size)

            if flat_global_size < max_total_local_size:
                flat_local_size = flat_global_size
            else:
                flat_local_size = local_size_multiple * (
                    max_total_local_size // local_size_multiple
                )

            # Find product(virtual_local_size) == flat_local_size <= max_total_local_size
            # Note: it's ok if local size elements are greater
            # than the corresponding global size elements as long as it minimizes the total
            # number of skipped threads.
            virtual_local_size = find_local_size_decomposition(virtual_global_size, flat_local_size)
        elif prod(virtual_local_size) > max_total_local_size:
            raise VirtualSizeError(
                f"Requested local size is greater than the maximum {max_total_local_size}"
            )

        # Global and local sizes supported by CUDA or OpenCL restricted number of dimensions,
        # which may have limited size, so we need to pack our multidimensional sizes.

        virtual_grid_size = [
            min_blocks(gs, ls)
            for gs, ls in zip(virtual_global_size, virtual_local_size, strict=True)
        ]
        bounding_global_size = [
            grs * ls for grs, ls in zip(virtual_grid_size, virtual_local_size, strict=True)
        ]

        if prod(virtual_grid_size) > prod(max_num_groups):
            # Report the bounding size in reversed form so that it matches the provided
            # virtual global size.
            raise VirtualSizeError(
                f"Bounding global size {list(reversed(bounding_global_size))} is too large"
            )

        local_groups = ShapeGroups(virtual_local_size, max_local_sizes)
        grid_groups = ShapeGroups(virtual_grid_size, max_num_groups)

        # These can be different lenghts because of expansion into multiple dimensions
        # find_bounding_shape() does.
        real_local_size = local_groups.bounding_shape
        real_grid_size = grid_groups.bounding_shape

        diff = len(real_local_size) - len(real_grid_size)
        real_local_size = real_local_size + [1] * (-diff)
        real_grid_size = real_grid_size + [1] * diff

        # This function will be used to translate between internal column-major vdims
        # and user-supplied row-major vdims.

        vsize_modules = VsizeModules.from_shape_data(
            virtual_local_size=virtual_local_size,
            virtual_global_size=virtual_global_size,
            bounding_global_size=bounding_global_size,
            virtual_grid_size=virtual_grid_size,
            local_groups=local_groups,
            grid_groups=grid_groups,
        )

        # For testing purposes
        # (Note that these will have column-major order, same as real_global/local_size)
        self._virtual_local_size = virtual_local_size
        self._virtual_global_size = virtual_global_size
        self._bounding_global_size = bounding_global_size
        self._virtual_grid_size = virtual_grid_size

        self.real_local_size = tuple(real_local_size)
        self.real_global_size = tuple(
            gs * ls for gs, ls in zip(real_grid_size, real_local_size, strict=True)
        )
        self.vsize_modules = vsize_modules



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy

if TYPE_CHECKING:  # pragma: no cover
    from .adapter_base import BufferAdapter
    from .context import BoundDevice
    from .queue import Queue


class Buffer:
    """A memory buffer on device."""

    device: BoundDevice
    """Device on which this buffer is allocated."""

    @classmethod
    def allocate(cls, device: BoundDevice, size: int) -> Buffer:
        """
        Allocate a buffer of ``size`` bytes.

        :param device: the device on which this buffer will be allocated.
        :param size: the buffer's size in bytes.
        """
        buffer_adapter = device.context._context_adapter.allocate(device._device_adapter, size)  # noqa: SLF001
        return cls(device, buffer_adapter)

    def __init__(self, device: BoundDevice, buffer_adapter: BufferAdapter):
        self.device = device
        self._buffer_adapter = buffer_adapter

    @property
    def kernel_arg(self) -> Any:
        # Has to be a property since buffer_adapter can be externally updated
        # (e.g. if it's a virtual buffer)
        return self._buffer_adapter.kernel_arg

    @property
    def offset(self) -> int:
        """
        Offset of this buffer (in bytes) from the beginning
        of the physical allocation it resides in.
        """
        return self._buffer_adapter.offset

    @property
    def size(self) -> int:
        """The buffer's size (in bytes)."""
        return self._buffer_adapter.size

    def set(
        self,
        queue: Queue,
        buf: numpy.ndarray[Any, numpy.dtype[Any]] | Buffer,
        *,
        no_async: bool = False,
    ) -> None:
        """
        Copy the contents of the host array or another buffer to this buffer.

        :param queue: the queue to use for the transfer.
        :param buf: the source - ``numpy`` array or a :py:class:`Buffer` object.
        :param no_async: if `True`, the transfer blocks until completion.
        """
        if queue.device != self.device:
            raise ValueError(
                f"Mismatched devices: queue on device {queue.device}, "
                f"buffer on device {self.device}"
            )

        buf_adapter: numpy.ndarray[Any, numpy.dtype[Any]] | BufferAdapter
        if isinstance(buf, numpy.ndarray):
            buf_adapter = numpy.ascontiguousarray(buf)
        elif isinstance(buf, Buffer):
            buf_adapter = buf._buffer_adapter  # noqa: SLF001
        else:
            raise TypeError(f"Cannot set from an object of type {type(buf)}")

        self._buffer_adapter.set(queue._queue_adapter, buf_adapter, no_async=no_async)  # noqa: SLF001

    def get(
        self,
        queue: Queue,
        host_array: numpy.ndarray[Any, numpy.dtype[Any]],
        *,
        async_: bool = False,
    ) -> None:
        """
        Copy the contents of the buffer to the host array.

        :param queue: the queue to use for the transfer.
        :param host_array: the destination array.
        :param async_: if `True`, the transfer is performed asynchronously.
        """
        if queue.device != self.device:
            raise ValueError(
                f"Mismatched devices: queue on device {queue.device}, "
                f"buffer on device {self.device}"
            )

        self._buffer_adapter.get(queue._queue_adapter, host_array, async_=async_)  # noqa: SLF001

    def get_sub_region(self, origin: int, size: int) -> Buffer:
        """
        Return a buffer object describing a subregion of this buffer.

        :param origin: the offset of the subregion.
        :param size: the size of the subregion.
        """
        if origin + size > self.size:
            raise ValueError("The requested subregion extends beyond the buffer length")
        return Buffer(self.device, self._buffer_adapter.get_sub_region(origin, size))



import pytest

import grunnur
from grunnur import cuda_api_id, opencl_api_id
from grunnur.testing import MockBackendFactory, MockPyCUDA, MockPyOpenCL


def test_mock_backend_factory(monkeypatch):
    fac = MockBackendFactory(monkeypatch)

    backend = fac.mock(opencl_api_id())
    assert isinstance(backend, MockPyOpenCL)
    assert grunnur.adapter_opencl.pyopencl._backend is backend

    fac.disable(opencl_api_id())
    assert grunnur.adapter_opencl.pyopencl is None

    backend = fac.mock(cuda_api_id())
    assert isinstance(backend, MockPyCUDA)
    assert grunnur.adapter_cuda.pycuda_driver is backend.pycuda_driver
    assert grunnur.adapter_cuda.pycuda_compiler is backend.pycuda_compiler

    fac.disable(cuda_api_id())
    assert grunnur.adapter_cuda.pycuda_driver is None
    assert grunnur.adapter_cuda.pycuda_compiler is None

    with pytest.raises(ValueError, match="Unknown API ID: foo"):
        fac.mock("foo")

    with pytest.raises(ValueError, match="Unknown API ID: foo"):
        fac.disable("foo")



# Arguments are actually used in templates, but Ruff cannot parse it.
# ruff: noqa: ARG005

import numpy
import pytest
from mako.template import Template as MakoTemplate

from grunnur.template import DefTemplate, RenderError, Template, _extract_def_source


def test_extract_def_source():
    src = """
        <%def name="add(varname)">
        ${varname} + ${num}
        </%def>

        <%def name='sub(varname, kwd=">")'>
        ${varname} - ${num}
        </%def>
    """

    def_src = _extract_def_source(src, "add")
    template = MakoTemplate(def_src)
    assert "add" in template.list_defs()
    assert "sub" not in template.list_defs()

    def_src = _extract_def_source(src, "sub")
    template = MakoTemplate(def_src)
    assert "sub" in template.list_defs()
    assert "add" not in template.list_defs()


def test_extract_def_source_missing_def():
    src = """
        <%def name="add(varname)">
        ${varname} + ${num}
        </%def>
    """

    # Technically, this warning and the subsequent return of the full source should occur
    # only if _extract_def_source() encounters a def it cannot recognize, but Mako still can -
    # otherwise get_def() will fail anyway.
    # But if I could think of something that would confuse _extract_def_source(), but
    # not Mako, I would modify _extract_def_source() accordingly.
    # So to model the situtation I'm just trying to find a non-existent def.

    with pytest.warns(SyntaxWarning):
        def_src = _extract_def_source(src, "sub")
    assert def_src == src


def test_template_from_associated_file():
    template = Template.from_associated_file(__file__)
    assert template.get_def("test").render().strip() == "template body"


def test_template_from_string():
    src = """
        <%def name="test()">
        template body
        </%def>
        """
    template = Template.from_string(src)
    assert template.get_def("test").render().strip() == "template body"


def test_missing_def():
    src = """
        <%def name="test()">
        template body
        </%def>
        """
    template = Template.from_string(src)
    with pytest.raises(AttributeError, match="Template has no def 'foo'"):
        template.get_def("foo")


def test_template_caching():
    src = """
        <%def name="test()">
        template body
        </%def>
        """
    template = Template.from_string(src)
    def1 = template.get_def("test")
    def2 = template.get_def("test")
    assert def1 is def2


def test_def_template_from_callable():
    template = DefTemplate.from_callable("test", lambda x, y: "${x} + ${y}")
    assert template.render(1, 2).strip() == "1 + 2"


def test_def_template_from_string():
    template = DefTemplate.from_string("test", ["x", "y"], "${x} + ${y}")
    assert template.render(1, 2).strip() == "1 + 2"


def test_render_error():
    template = DefTemplate.from_callable("test", lambda x, y: "${x} + ${y} + ${z}")
    with pytest.raises(RenderError) as e:
        template.render(1, 2, kwd=3)
    assert e.value.args == (1, 2)
    assert e.value.globals == dict(kwd=3)
    assert e.value.source == template.source
    assert type(e.value.exception) is NameError
    assert str(e.value.exception) in str(e.value)


def test_render_error_pass_through():
    # check that if one template tries to render another, the inner RenderError gets propagated
    template1 = DefTemplate.from_callable("test1", lambda x, y: "${x} + ${y} + ${z}")
    template2 = DefTemplate.from_callable("test2", lambda a, t1: "${a} + ${t1.render(a, 1)}")
    with pytest.raises(RenderError) as e:
        template2.render(10, template1, kwd=3)
    assert e.value.args == (10, 1)
    assert e.value.globals == dict()
    assert e.value.source == template1.source
    assert type(e.value.exception) is NameError
    assert str(e.value.exception) in str(e.value)


def test_template_builtins():
    # Check for the builtins we add to every template
    template = DefTemplate.from_callable("test", lambda numpy_ref: "${numpy == numpy_ref}")
    assert template.render(numpy).strip() == "True"

    template = DefTemplate.from_string("test", ["numpy_ref"], "${numpy == numpy_ref}")
    assert template.render(numpy).strip() == "True"

    template = Template.from_string('<%def name="test(numpy_ref)">${numpy == numpy_ref}</%def>')
    template_def = template.get_def("test")
    assert template_def.render(numpy).strip() == "True"

    template = Template.from_associated_file(__file__)
    assert template.get_def("test_builtins").render(numpy).strip() == "True"



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from typing import TYPE_CHECKING, Any, overload

from .api import API, cuda_api_id
from .device import Device, DeviceFilter
from .device_discovery import select_devices
from .platform import Platform, PlatformFilter
from .utils import all_same, normalize_object_sequence

if TYPE_CHECKING:  # pragma: no cover
    from .adapter_base import ContextAdapter, DeviceAdapter


class BoundDevice(Device):
    """A :py:class:`~grunnur.Device` object in a :py:class:`~grunnur.Context`."""

    context: Context
    """The context this device belongs to."""

    def __init__(self, context: Context, device_adapter: DeviceAdapter):
        super().__init__(device_adapter)
        self.context = context

        # A proper hashing would require `Context` to be hashable too,
        # but `BoundDevice` objects are only ever used in small collections
        # and with all the device indices being different.
        # If somehow there's a hash collision, it will be taken care of by ``__eq__``.
        self._hash = hash(device_adapter)

    def as_unbound(self) -> Device:
        """
        :meta private:
        Returns the unbound :py:class:`Device` object.
        """
        return Device(self._device_adapter)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, BoundDevice)
            and self.context == other.context
            and super().__eq__(other)
        )

    def __hash__(self) -> int:
        return self._hash

    def __str__(self) -> str:
        return super().__str__() + " in " + str(self.context)


class BoundMultiDevice(Sequence[BoundDevice]):
    """A sequence of bound devices belonging to the same context."""

    context: Context
    """The context these devices belong to."""

    @classmethod
    def from_bound_devices(cls, devices: Sequence[BoundDevice]) -> BoundMultiDevice:
        """
        Creates this object from a sequence of bound devices
        (note that a ``BoundMultiDevice`` object itself can serve as such a sequence).
        """
        if not all_same(device.context for device in devices):
            raise ValueError("All devices in a multi-device must belong to the same context")

        if len(set(devices)) != len(devices):
            raise ValueError("All devices in a multi-device must be distinct")

        return cls(devices[0].context, [device._device_adapter for device in devices])  # noqa: SLF001

    def __init__(self, context: Context, device_adapters: Sequence[DeviceAdapter]):
        self.context = context
        self._devices = [BoundDevice(context, device_adapter) for device_adapter in device_adapters]
        self._devices_as_set = set(self._devices)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, BoundMultiDevice)
            and self.context == other.context
            and self._devices == other._devices
        )

    def issubset(self, devices: BoundMultiDevice) -> bool:
        return self._devices_as_set.issubset(devices._devices_as_set)  # noqa: SLF001

    def __iter__(self) -> Iterator[BoundDevice]:
        return iter(self._devices)

    @overload
    def __getitem__(self, idx: int) -> BoundDevice: ...

    @overload
    def __getitem__(self, idx: slice | Iterable[int]) -> BoundMultiDevice: ...

    def __getitem__(self, idx: int | slice | Iterable[int]) -> BoundDevice | BoundMultiDevice:
        """
        Given a single index, returns a single :py:class:`BoundDevice`.
        Given a sequence of indices, returns a :py:class:`BoundMultiDevice` object
        containing respective devices.

        The indices correspond to the list of devices used to create this context.
        """
        if isinstance(idx, Iterable):
            return BoundMultiDevice.from_bound_devices([self._devices[i] for i in idx])
        if isinstance(idx, slice):
            return BoundMultiDevice.from_bound_devices(self._devices[idx])
        return self._devices[idx]

    def __len__(self) -> int:
        return len(self._devices)


class Context:
    """GPGPU context."""

    platform: Platform
    """The platform this context is based on."""

    api: API
    """The API this context is based on."""

    @classmethod
    def from_devices(cls, devices: Sequence[Device]) -> Context:
        """
        Creates a context from a device or an iterable of devices.

        :param devices: one or several devices to use.
        """
        devices = normalize_object_sequence(devices, Device)

        platforms = [device.platform for device in devices]
        if not all_same(platforms):
            raise ValueError("All devices must belong to the same platform")
        platform = platforms[0]

        device_adapters = [device._device_adapter for device in devices]  # noqa: SLF001
        api_adapter = platform.api._api_adapter  # noqa: SLF001
        context_adapter = api_adapter.make_context_adapter_from_device_adapters(device_adapters)

        return cls(context_adapter)

    @classmethod
    def from_backend_devices(cls, backend_devices: Sequence[Any]) -> Context:
        """Creates a context from a single or several backend device objects."""
        devices = [Device.from_backend_device(backend_device) for backend_device in backend_devices]
        return cls.from_devices(devices)

    @classmethod
    def from_backend_contexts(
        cls, backend_contexts: Sequence[Any], *, take_ownership: bool = False
    ) -> Context:
        """
        Creates a context from a single or several backend device contexts.
        If ``take_ownership`` is ``True``, this object will be responsible for the lifetime
        of backend context objects (only important for the CUDA backend).
        """
        for api in API.all_available():
            if api._api_adapter.isa_backend_context(backend_contexts[0]):  # noqa: SLF001
                context_adapter = api._api_adapter.make_context_adapter_from_backend_contexts(  # noqa: SLF001
                    backend_contexts, take_ownership=take_ownership
                )
                return cls(context_adapter)
        raise TypeError(
            f"{type(backend_contexts[0])} objects were not recognized as contexts by any API"
        )

    @classmethod
    def from_criteria(
        cls,
        api: API,
        *,
        interactive: bool = False,
        devices_num: int | None = 1,
        device_filter: DeviceFilter | None = None,
        platform_filter: PlatformFilter | None = None,
    ) -> Context:
        """
        Finds devices matching the given criteria and creates a
        :py:class:`Context` object out of them.

        :param interactive: passed to :py:func:`select_devices`.
        :param devices_num: passed to :py:func:`select_devices` as ``quantity``.
        :param device_filters: passed to :py:func:`select_devices`.
        """
        devices = select_devices(
            api,
            interactive=interactive,
            quantity=devices_num,
            device_filter=device_filter,
            platform_filter=platform_filter,
        )
        return cls.from_devices(devices)

    def __init__(self, context_adapter: ContextAdapter):
        self._context_adapter = context_adapter
        self._device_adapters = context_adapter.device_adapters
        self.platform = Platform(next(iter(self._device_adapters.values())).platform_adapter)
        self.api = self.platform.api

    @property
    def devices(self) -> BoundMultiDevice:
        """
        Returns the :py:class:`~grunnur.context.BoundMultiDevice`
        encompassing all the devices in this context.
        """
        # Need to create it on-demand to avoid a circular reference.
        device_adapters = [
            self._device_adapters[device_idx] for device_idx in self._context_adapter.device_order
        ]
        return BoundMultiDevice(self, device_adapters)

    @property
    def device(self) -> BoundDevice:
        if len(self._device_adapters) > 1:
            raise RuntimeError("The `device` shortcut only works for single-device contexts")

        return self.devices[0]

    def deactivate(self) -> None:
        """
        For CUDA API: deactivates this context, popping all the CUDA context objects from the stack.
        Other APIs: no effect.

        Only call it if you need to manage CUDA contexts manually,
        and created this object with `take_ownership = False`.
        If `take_ownership = True` contexts will be deactivated automatically in the destructor.
        """
        self._context_adapter.deactivate()



import weakref
from collections import Counter

import numpy
import pytest

from grunnur import Array, Buffer, Context, MultiQueue, Program, Queue, dtypes
from grunnur.adapter_base import APIID
from grunnur.virtual_alloc import TrivialManager, ZeroOffsetManager, extract_dependencies


def allocate_test_set(virtual_alloc, allocate_callable):
    # Allocate virtual buffers with dependencies

    buffers_metadata = [
        (0, 2, []),
        (1, 5, [0]),
        (2, 5, [0]),
        (3, 10, [1, 2]),
        (4, 8, [1, 3]),
        (5, 5, []),
        (6, 3, [3]),
        (7, 10, [2, 5]),
        (8, 5, [1, 6]),
        (9, 20, [2, 4, 8]),
    ]

    buffers = {}
    for name, size, deps in buffers_metadata:
        allocator = virtual_alloc.allocator([buffers[d] for d in deps])
        buffers[name] = allocate_callable(allocator, size)

    return buffers_metadata, buffers


@pytest.mark.parametrize("pack", [False, True], ids=["no_pack", "pack"])
def test_contract(context, valloc_cls, pack):
    dtype = numpy.int32

    program = Program(
        [context.device],
        """
        KERNEL void fill(GLOBAL_MEM ${ctype} *dest, ${ctype} val)
        {
            const SIZE_T i = get_global_id(0);
            dest[i] = val;
        }
        """,
        render_globals=dict(ctype=dtypes.ctype(dtype)),
    )
    fill = program.kernel.fill

    queue = Queue(context.device)
    virtual_alloc = valloc_cls(context.device)

    buffers_metadata, arrays = allocate_test_set(
        virtual_alloc,
        # Bump size to make sure buffer alignment doesn't hide any out-of-bounds access
        lambda allocator, size: Array.empty(
            context.device, [size * 100], dtype, allocator=allocator
        ),
    )
    dependencies = {id_: deps for id_, _, deps in buffers_metadata}

    if pack:
        virtual_alloc.pack(queue)

    # Clear all arrays
    for name in sorted(arrays.keys()):
        fill(queue, arrays[name].shape, None, arrays[name], dtype(0))

    for i, name in enumerate(sorted(arrays.keys())):
        val = dtype(i + 1)
        fill(queue, arrays[name].shape, None, arrays[name], val)
        # According to the virtual allocator contract, the allocated buffer
        # will not intersect with the buffers from the specified dependencies.
        # So we're filling the buffer and checking that the dependencies did not change.
        for dep in dependencies[name]:
            assert (arrays[dep].get(queue) != val).all()


@pytest.mark.parametrize("pack", [False, True], ids=["no_pack", "pack"])
def test_contract_mocked(mock_backend_pycuda, mock_context_pycuda, valloc_cls, pack):
    # Using PyCUDA backend here because it tracks the allocations.

    context = mock_context_pycuda
    queue = Queue(context.device)
    virtual_alloc = valloc_cls(context.device)

    buffers_metadata, buffers = allocate_test_set(
        virtual_alloc, lambda allocator, size: allocator(context.device, size)
    )

    for name, size, _deps in buffers_metadata:
        # Virtual buffer size should be exactly as requested
        assert buffers[name].size == size
        # The real buffer behind the virtual buffer may be larger
        assert buffers[name]._buffer_adapter._real_buffer_adapter.size >= size

    if pack:
        virtual_alloc.pack(queue)

    for i, metadata in enumerate(buffers_metadata):
        name, _size, deps = metadata
        buffers[name].set(queue, numpy.ones(buffers[name].size, numpy.uint8) * i, no_async=True)
        # According to the virtual allocator contract, the allocated buffer
        # will not intersect with the buffers from the specified dependencies.
        # So we're filling the buffer and checking that the dependencies did not change.
        for dep in deps:
            arr = numpy.zeros(buffers[dep].size, numpy.uint8)
            buffers[dep].get(queue, arr, async_=False)
            assert (arr != i).all()

    # Check that after deleting virtual buffers all the real buffers are freed as well
    del buffers
    assert mock_backend_pycuda.allocation_count() == 0


def test_extract_dependencies(mock_context):
    virtual_alloc = TrivialManager(mock_context.device).allocator()

    vbuf = virtual_alloc(mock_context.device, 100)
    varr = Array.empty(mock_context.device, [100], numpy.int32, allocator=virtual_alloc)

    assert extract_dependencies(vbuf) == {vbuf._buffer_adapter._id}
    assert extract_dependencies(varr) == {varr.data._buffer_adapter._id}
    assert extract_dependencies([vbuf, varr]) == {
        vbuf._buffer_adapter._id,
        varr.data._buffer_adapter._id,
    }

    class DependencyHolder:
        __virtual_allocations__ = (vbuf, varr)

    assert extract_dependencies(DependencyHolder()) == {
        vbuf._buffer_adapter._id,
        varr.data._buffer_adapter._id,
    }

    # An object not having any dependencies
    assert extract_dependencies(123) == set()


def check_statistics(buffers_metadata, stats):
    virtual_sizes = [size for _, size, _ in buffers_metadata]
    assert stats.virtual_size_total == sum(virtual_sizes)
    assert stats.virtual_num == len(virtual_sizes)
    assert stats.virtual_sizes == dict(Counter(virtual_sizes))

    assert stats.real_size_total <= sum(virtual_sizes)
    assert stats.real_num <= len(virtual_sizes)


def test_statistics(mock_context, valloc_cls):
    context = mock_context
    queue = Queue(context.device)
    virtual_alloc = valloc_cls(context.device)

    buffers_metadata, buffers = allocate_test_set(
        virtual_alloc, lambda allocator, size: allocator(context.device, size)
    )

    stats = virtual_alloc.statistics()
    check_statistics(buffers_metadata, stats)

    virtual_alloc.pack(queue)

    stats = virtual_alloc.statistics()
    check_statistics(buffers_metadata, stats)

    s = str(stats)
    assert str(stats.real_size_total) in s
    assert str(stats.real_num) in s
    assert str(stats.virtual_size_total) in s
    assert str(stats.virtual_num) in s


def test_non_existent_dependencies(mock_context, valloc_cls):
    context = mock_context
    virtual_alloc = valloc_cls(context.device)
    with pytest.raises(ValueError, match="12345"):
        virtual_alloc._allocate_virtual(100, {12345})


def test_virtual_buffer(mock_4_device_context_pyopencl):
    # Using an OpenCL mock context here because it keeps track of buffer migrations

    context = mock_4_device_context_pyopencl
    mqueue = MultiQueue.on_devices(context.devices)
    dev0 = context.devices[0]

    virtual_alloc = TrivialManager(dev0)

    allocator = virtual_alloc.allocator()
    vbuf = allocator(dev0, 100)

    assert vbuf.size == 100
    assert isinstance(vbuf.kernel_arg._buffer, bytes)
    assert vbuf.device == dev0
    with pytest.raises(NotImplementedError):
        vbuf.get_sub_region(0, 50)
    assert vbuf.offset == 0

    arr = numpy.arange(100).astype(numpy.uint8)
    vbuf.set(mqueue.queues[dev0], arr)
    res = numpy.empty_like(arr)
    vbuf.get(mqueue.queues[dev0], res)
    assert (arr == res).all()


def test_virtual_alloc_device_check(mock_4_device_context_pyopencl):
    context = mock_4_device_context_pyopencl
    virtual_alloc = TrivialManager(context.devices[0])
    alloc = virtual_alloc.allocator()
    with pytest.raises(ValueError, match="This allocator is attached to device"):
        alloc(context.devices[1], 100)



import numpy
import pytest

from grunnur import API, Buffer, Context, Queue, cuda_api_id, opencl_api_id


def test_allocate_and_copy(mock_or_real_context):  # noqa: PLR0915
    context, mocked = mock_or_real_context

    length = 100
    dtype = numpy.dtype("int32")
    size = length * dtype.itemsize

    arr = numpy.arange(length).astype(dtype)

    buf = Buffer.allocate(context.device, size)
    assert buf.size == size
    assert buf.offset == 0

    # Just covering the existence of the attribute.
    # Hard to actually check it without running a kernel
    assert buf.kernel_arg is not None

    queue = Queue(context.device)
    buf.set(queue, arr)

    # Read the whole buffer
    res = numpy.empty_like(arr)
    buf.get(queue, res)
    queue.synchronize()
    assert (res == arr).all()

    # Read a subregion
    buf_region = buf.get_sub_region(25 * dtype.itemsize, 50 * dtype.itemsize)
    arr_region = arr[25 : 25 + 50]
    res_region = numpy.empty_like(arr_region)
    buf_region.get(queue, res_region)
    queue.synchronize()
    assert (res_region == arr_region).all()

    # Check that our mock can detect taking a sub-region of a sub-region (segfault in OpenCL)
    if mocked and context.api.id == opencl_api_id():
        with pytest.raises(RuntimeError, match="Cannot create a subregion of subregion"):
            buf_region.get_sub_region(0, 10)

    # Write a subregion
    arr_region = (numpy.ones(50) * 100).astype(dtype)
    arr[25 : 25 + 50] = arr_region
    buf_region.set(queue, arr_region)
    buf.get(queue, res)
    queue.synchronize()
    assert (res == arr).all()

    # Subregion of subregion
    if context.api.id == cuda_api_id():
        # In OpenCL that leads to segfault, but with CUDA we just emulate that with pointers.
        arr_region2 = (numpy.ones(20) * 200).astype(dtype)
        arr[25 + 20 : 25 + 40] = arr_region2
        buf_region2 = buf_region.get_sub_region(20 * dtype.itemsize, 20 * dtype.itemsize)
        buf_region2.set(queue, arr_region2)
        buf.get(queue, res)
        queue.synchronize()
        assert (res == arr).all()

    # Device-to-device copy
    buf2 = Buffer.allocate(context.device, size * 2)
    buf2.set(queue, numpy.ones(length * 2, dtype))
    buf2_view = buf2.get_sub_region(50 * dtype.itemsize, 100 * dtype.itemsize)
    buf2_view.set(queue, buf)
    res2 = numpy.empty(length * 2, dtype)
    buf2.get(queue, res2)
    queue.synchronize()
    assert (res2[50:150] == arr).all()
    assert (res2[:50] == 1).all()
    assert (res2[150:] == 1).all()

    # Device-to-device copy (no_async)
    buf2 = Buffer.allocate(context.device, size * 2)
    buf2.set(queue, numpy.ones(length * 2, dtype))
    buf2_view = buf2.get_sub_region(50 * dtype.itemsize, 100 * dtype.itemsize)
    buf2_view.set(queue, buf, no_async=True)
    res2 = numpy.empty(length * 2, dtype)
    buf2.get(queue, res2)
    queue.synchronize()
    assert (res2[50:150] == arr).all()
    assert (res2[:50] == 1).all()
    assert (res2[150:] == 1).all()


def test_subregion_overflow(mock_context):
    buf = Buffer.allocate(mock_context.device, 100)
    with pytest.raises(
        ValueError, match="The requested subregion extends beyond the buffer length"
    ):
        buf.get_sub_region(50, 70)


def test_migrate_on_copy(monkeypatch, mock_4_device_context):
    context = mock_4_device_context
    size = 100
    src = Buffer.allocate(context.devices[0], size)
    dest = Buffer.allocate(context.devices[0], size)

    queue0 = Queue(context.devices[0])
    queue1 = Queue(context.devices[1])

    arr = numpy.arange(size).astype(numpy.uint8)

    src.set(queue0, arr)
    queue0.synchronize()

    monkeypatch.setattr(queue1, "device", queue0.device)

    if context.api.id == cuda_api_id():
        message = "Trying to access an allocation from a device different to where it was created"
    else:
        message = "Trying to access a buffer from a device different to where it was migrated to"

    with pytest.raises(RuntimeError, match=message):
        dest.set(queue1, src)


def test_flags(mock_backend_pyopencl):
    backend = mock_backend_pyopencl

    normal_flags = backend.pyopencl.mem_flags.READ_WRITE
    special_flags = normal_flags | backend.pyopencl.mem_flags.ALLOC_HOST_PTR

    backend.add_platform_with_devices("Apple", ["GeForce", "Foo", "Bar"])
    backend.add_platform_with_devices("Baz", ["GeForce", "Foo", "Bar"])

    # Multi-device on Apple platform with one of the devices being GeForce:
    # need special Buffer flags
    api = API.from_api_id(backend.api_id)
    context = Context.from_devices([api.platforms[0].devices[0], api.platforms[0].devices[1]])
    buf = Buffer.allocate(context.devices[0], 100)
    assert buf._buffer_adapter._pyopencl_buffer.flags == special_flags

    # None of the devices is GeForce
    context = Context.from_devices([api.platforms[0].devices[1], api.platforms[0].devices[2]])
    buf = Buffer.allocate(context.devices[0], 100)
    assert buf._buffer_adapter._pyopencl_buffer.flags == normal_flags

    # Only one device
    context = Context.from_devices([api.platforms[0].devices[0]])
    buf = Buffer.allocate(context.device, 100)
    assert buf._buffer_adapter._pyopencl_buffer.flags == normal_flags

    # Not an Apple platform
    context = Context.from_devices([api.platforms[1].devices[0], api.platforms[1].devices[1]])
    buf = Buffer.allocate(context.devices[0], 100)
    assert buf._buffer_adapter._pyopencl_buffer.flags == normal_flags


def test_set_from_wrong_type(mock_context):
    buf = Buffer.allocate(mock_context.device, 100)
    queue = Queue(mock_context.device)
    with pytest.raises(TypeError, match="Cannot set from an object of type <class 'int'>"):
        buf.set(queue, 1)


def test_mismatched_devices(mock_4_device_context):
    context = mock_4_device_context
    buf = Buffer.allocate(context.devices[0], 100)
    queue = Queue(context.devices[1])
    arr = numpy.ones(100, numpy.uint8)

    with pytest.raises(ValueError, match="Mismatched devices: queue on device"):
        buf.get(queue, arr)

    with pytest.raises(ValueError, match="Mismatched devices: queue on device"):
        buf.set(queue, arr)



# This is supposed to be used in CLI, so the functions contain printing.
# ruff: noqa: T201

from __future__ import annotations

from typing import TYPE_CHECKING

from .device import Device, DeviceFilter
from .platform import Platform, PlatformFilter

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

    from .api import API


def platforms_and_devices_by_mask(
    api: API,
    quantity: int | None = 1,
    device_filter: DeviceFilter | None = None,
    platform_filter: PlatformFilter | None = None,
) -> list[tuple[Platform, list[Device]]]:
    """
    Returns all tuples (platform, list of devices) where the platform name and device names
    satisfy the given criteria, and there are at least ``quantity`` devices in the list.
    """
    results = []

    suitable_platforms = Platform.all_filtered(api, platform_filter)

    for platform in suitable_platforms:
        suitable_devices = Device.all_filtered(platform, device_filter)

        if (quantity is None and len(suitable_devices) > 0) or (
            quantity is not None and len(suitable_devices) >= quantity
        ):
            results.append((platform, suitable_devices))

    return results


def _select_devices_interactive(
    suitable_pds: Sequence[tuple[Platform, Sequence[Device]]], quantity: int | None = 1
) -> list[Device]:
    if len(suitable_pds) == 1:
        platform, devices = suitable_pds[0]
        print(f"Platform: {platform.name}")
    else:
        print("Platforms:")
        for pnum, pd in enumerate(suitable_pds):
            platform, _ = pd
            print(f"[{pnum}]: {platform.name}")

        default_pnum = 0
        selected_pnum = default_pnum
        print(f"Choose the platform [{default_pnum}]: ", end="")
        pnum_input = input()
        if pnum_input != "":
            selected_pnum = int(pnum_input)

        platform, devices = suitable_pds[selected_pnum]

    if quantity is None or (quantity is not None and len(devices) == quantity):
        selected_devices = list(devices)
        print(f"Devices: {[device.name for device in selected_devices]}")
    else:
        print("Devices:")
        default_dnums = list(range(quantity))
        for dnum, device in enumerate(devices):
            print(f"[{dnum}]: {device.name}")
        default_dnums_str = ", ".join(str(dnum) for dnum in default_dnums)
        print(f"Choose the device(s), comma-separated [{default_dnums_str}]: ", end="")
        selected_dnums = default_dnums
        dnums_input = input()
        if dnums_input != "":
            selected_dnums = [int(dnum) for dnum in dnums_input.split(",")]
            if len(selected_dnums) != quantity:
                raise ValueError(f"Exactly {quantity} devices must be selected")

        selected_devices = [devices[dnum] for dnum in selected_dnums]

    return selected_devices


def select_devices(
    api: API,
    *,
    interactive: bool = False,
    quantity: int | None = 1,
    device_filter: DeviceFilter | None = None,
    platform_filter: PlatformFilter | None = None,
) -> list[Device]:
    """
    Using the results from :py:func:`platforms_and_devices_by_mask`, either lets the user
    select the devices (from the ones matching the criteria) interactively,
    or takes the first matching list of ``quantity`` devices.

    :param interactive: if ``True``, shows a dialog to select the devices.
        If ``False``, selects the first matching ones.
    :param quantity: passed to :py:func:`platforms_and_devices_by_mask`.
    :param device_filters: passed to :py:func:`platforms_and_devices_by_mask`.
    """
    suitable_pds = platforms_and_devices_by_mask(
        api, quantity, device_filter=device_filter, platform_filter=platform_filter
    )

    if len(suitable_pds) == 0:
        quantity_val = "any" if quantity is None else quantity
        raise ValueError(
            f"Could not find {quantity_val} devices on a single platform "
            "matching the given criteria"
        )

    if interactive:
        return _select_devices_interactive(suitable_pds, quantity=quantity)

    _, devices = suitable_pds[0]
    return devices if quantity is None else devices[:quantity]



import pytest

from grunnur import API, Device, DeviceFilter, Platform, cuda_api_id, opencl_api_id
from grunnur.adapter_base import DeviceType
from grunnur.testing import PyCUDADeviceInfo, PyOpenCLDeviceInfo


def test_all(mock_backend):
    mock_backend.add_devices(["Device2", "Device3"])
    api = API.from_api_id(mock_backend.api_id)
    platform = Platform.from_index(api, 0)
    devices = Device.all(platform)
    device_names = {device.name for device in devices}
    assert device_names == {"Device2", "Device3"}


@pytest.mark.parametrize("unique_only", [False, True], ids=["all", "unique"])
def test_all_filtered(mock_backend, unique_only):
    mock_backend.add_devices(["foo-bar", "foo-baz", "bar-baz", "foo-baz"])
    api = API.from_api_id(mock_backend.api_id)
    platform = Platform.from_index(api, 0)
    devices = Device.all_filtered(
        platform,
        DeviceFilter(include_masks=["foo"], exclude_masks=["bar"], unique_only=unique_only),
    )
    if unique_only:
        assert len(devices) == 1
        assert devices[0].name == "foo-baz"
    else:
        assert len(devices) == 2
        assert devices[0].name == "foo-baz"
        assert devices[1].name == "foo-baz"
        assert devices[0] != devices[1]


def test_from_backend_device(mock_backend):
    mock_backend.add_devices(["Device1"])

    api = API.from_api_id(mock_backend.api_id)

    if api.id == opencl_api_id():
        backend_device = mock_backend.pyopencl.get_platforms()[0].get_devices()[0]
    elif api.id == cuda_api_id():
        backend_device = mock_backend.pycuda_driver.Device(0)
    else:
        raise NotImplementedError

    with pytest.raises(TypeError, match="was not recognized as a device object"):
        Device.from_backend_device(1)

    device = Device.from_backend_device(backend_device)
    assert device.platform.api == api
    if api.id != cuda_api_id():
        assert device.platform.name == "Platform0"
    assert device.name == "Device1"


def test_from_index(mock_backend):
    mock_backend.add_devices(["Device1", "Device2"])
    api = API.from_api_id(mock_backend.api_id)
    platform = Platform.from_index(api, 0)
    device = Device.from_index(platform, 1)
    assert device.name == "Device2"


def test_params(mock_backend):
    mock_backend.add_devices(["Device1"])
    api = API.from_api_id(mock_backend.api_id)
    platform = Platform.from_index(api, 0)
    device = Device.from_index(platform, 0)

    params1 = device.params
    params2 = device.params
    assert params1 is params2  # check caching


def test_eq(mock_backend):
    mock_backend.add_devices(["Device0", "Device1"])
    api = API.from_api_id(mock_backend.api_id)

    platform = Platform.from_index(api, 0)
    d0_v1 = Device.from_index(platform, 0)
    d0_v2 = Device.from_index(platform, 0)
    d1 = Device.from_index(platform, 1)

    assert d0_v1 is not d0_v2
    assert d0_v1 == d0_v2
    assert d0_v1 != d1


def test_hash(mock_backend):
    mock_backend.add_devices(["Device0", "Device1"])
    api = API.from_api_id(mock_backend.api_id)

    platform = Platform.from_index(api, 0)
    d0 = Device.from_index(platform, 0)
    d1 = Device.from_index(platform, 1)

    d = {d0: 0, d1: 1}
    assert d[d0] == 0
    assert d[d1] == 1


def test_attributes(mock_backend):
    mock_backend.add_devices(["Device1"])
    api = API.from_api_id(mock_backend.api_id)
    p = Platform.from_index(api, 0)
    d = Device.from_index(p, 0)

    assert d.platform == p
    assert d.name == "Device1"
    assert d.shortcut == p.shortcut + ",0"
    assert str(d) == "device(" + d.shortcut + ")"


def test_device_parameters_opencl(mock_backend_pyopencl):
    device_info = PyOpenCLDeviceInfo(
        type_=DeviceType.GPU,
        max_work_group_size=512,
        max_work_item_sizes=[512, 512, 512],
        local_mem_size=32 * 1024,
        address_bits=32,
        max_compute_units=4,
    )

    mock_backend_pyopencl.add_devices([device_info])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.type == device_info.type
    assert d.params.max_total_local_size == device_info.max_work_group_size
    assert d.params.max_local_sizes == tuple(device_info.max_work_item_sizes)
    assert d.params.warp_size == 32
    mng = 2**device_info.address_bits // device_info.max_work_group_size
    assert d.params.max_num_groups == (mng,) * len(device_info.max_work_item_sizes)
    assert d.params.local_mem_size == device_info.local_mem_size
    assert d.params.local_mem_banks == 32
    assert d.params.compute_units == device_info.max_compute_units
    assert d.params.align_words(4) >= 1
    assert d.params.align_words(8) >= 1
    assert d.params.align_words(16) >= 1


def test_device_parameters_opencl_apple_cpu(mock_backend_pyopencl):
    device_info = PyOpenCLDeviceInfo(type_=DeviceType.CPU, max_work_group_size=512)

    mock_backend_pyopencl.add_platform_with_devices("Apple", [device_info])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.max_total_local_size == 1
    assert d.params.max_local_sizes == (1, 1, 1)


def test_device_parameters_opencl_cpu(mock_backend_pyopencl):
    device_info = PyOpenCLDeviceInfo(type_=DeviceType.CPU)

    mock_backend_pyopencl.add_devices([device_info])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.local_mem_banks == 1
    assert d.params.warp_size == 1


def test_device_parameters_opencl_cuda1(mock_backend_pyopencl):
    # CUDA 1.x

    device_info = PyOpenCLDeviceInfo(
        extensions=["cl_nv_device_attribute_query"], compute_capability_major_nv=1, warp_size_nv=7
    )

    mock_backend_pyopencl.add_devices([device_info])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.local_mem_banks == 16
    assert d.params.warp_size == device_info.warp_size_nv


def test_device_parameters_opencl_cuda2(mock_backend_pyopencl):
    # CUDA 2.x+

    device_info = PyOpenCLDeviceInfo(
        extensions=["cl_nv_device_attribute_query"], compute_capability_major_nv=2, warp_size_nv=7
    )

    mock_backend_pyopencl.add_devices([device_info])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.local_mem_banks == 32
    assert d.params.warp_size == device_info.warp_size_nv


def test_device_parameters_opencl_unknown_nv(mock_backend_pyopencl):
    # Unknown nVidia device

    device_info = PyOpenCLDeviceInfo(vendor="NVIDIA")

    mock_backend_pyopencl.add_devices([device_info])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.local_mem_banks == 32
    assert d.params.warp_size == 32


def test_device_parameters_cuda(mock_backend_pycuda):
    device_info = PyCUDADeviceInfo(
        max_threads_per_block=512,
        max_block_dim_x=512,
        max_block_dim_y=512,
        max_block_dim_z=32,
        max_grid_dim_x=2**16,
        max_grid_dim_y=2**16,
        max_grid_dim_z=256,
        warp_size=7,
        max_shared_memory_per_block=58 * 1024,
        multiprocessor_count=3,
        compute_capability=10,
    )

    mock_backend_pycuda.add_devices([device_info])
    api = API.from_api_id(mock_backend_pycuda.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.type == DeviceType.GPU
    assert d.params.max_total_local_size == device_info.max_threads_per_block
    assert d.params.max_local_sizes == (
        device_info.max_block_dim_x,
        device_info.max_block_dim_y,
        device_info.max_block_dim_z,
    )
    assert d.params.warp_size == device_info.warp_size
    assert d.params.max_num_groups == (
        device_info.max_grid_dim_x,
        device_info.max_grid_dim_y,
        device_info.max_grid_dim_z,
    )
    assert d.params.local_mem_size == device_info.max_shared_memory_per_block
    assert d.params.local_mem_banks == 32
    assert d.params.compute_units == device_info.multiprocessor_count
    assert d.params.align_words(4) >= 1
    assert d.params.align_words(8) >= 1
    assert d.params.align_words(16) >= 1


def test_device_parameters_cuda_1(mock_backend_pycuda):
    # CUDA 1.x

    device_info = PyCUDADeviceInfo(compute_capability=1)

    mock_backend_pycuda.add_devices([device_info])
    api = API.from_api_id(mock_backend_pycuda.api_id)
    d = api.platforms[0].devices[0]

    assert d.params.local_mem_banks == 16



from collections.abc import Sequence
from typing import Any, Literal, Optional, overload

import numpy

from . import driver

class SourceModule:
    def __init__(
        self,
        source: str,
        *,
        options: Sequence[str] | None = None,
        keep: bool = False,
        no_extern_c: bool = False,
        cache_dir: str | None = None,
        include_dirs: Sequence[str] = [],
    ): ...
    def get_function(self, name: str) -> Function: ...
    def get_global(self, name: str) -> tuple[int, int]: ...

class Function:
    @overload
    def get_attribute(
        self, attr: Literal[driver.function_attribute.MAX_THREADS_PER_BLOCK]
    ) -> int: ...
    @overload
    def get_attribute(self, attr: driver.function_attribute) -> Any: ...
    def __call__(
        self,
        *args: driver.DeviceAllocation | numpy.generic,
        block: tuple[int, ...],
        grid: tuple[int, ...],
        stream: driver.Stream | None = None,
        shared: int = 0,
    ) -> None: ...
